MODULE ParnacDataSourceInit

    ! =============================================================================
    !
    !   Programmer: Koos Huijssen / Jasper de Koning
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    !   Copyright (C)   Laboratory of Electromagnetic Research
    !                   Delft University of Technology, Delft, The Netherlands
    !
    !   Permission to copy or distribute this software or documentation
    !   in hard copy or soft copy granted only by written license
    !   obtained from the Laboratory of Electromagnetic Research.
    !   All rights reserved. No parts of this publication may be
    !   reproduced, stored in a retrieval system (e.g. in memory, disk,
    !   or core) or be transmitted by any means, electronic, mechanical,
    !   photocopy, recording, or otherwise, without written permission
    !   from the publisher.
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The module ParnacDataSourceInit contains subroutines for the definition
    !   of the primary source.
    !
    ! *****************************************************************************
    !
    !   MODULES USED
    !
    !#ifdef FFTW_C_NAMING
        include "fftw_wrappers.inc"
    !#endif
    
        USE Types
        USE Constants
    #ifndef RUNESTIMATOR
        USE FFTW_cons
    #endif
    
        USE ParnacGeneral
        USE ParnacIdentifierDef
        USE ParnacDataDef
        USE ParnacParamDef
    
    !        USE ParnacDataSpaceInit, ONLY : &
        USE ParnacParamInit, ONLY: &
            BINARY_read_sourcedef, Binary_Read3DArraySize, Binary_Read3DArrayTDelay, &
            Binary_Read1DArray, Binary_Read1DArraySize, &
            ASCII_Read3DArray, ASCII_Read3DArraySize, &
            ASCII_Read1DArray, ASCII_Read1DArraySize
    #ifndef RUNESTIMATOR
        USE SpecialFun, ONLY: &
            CiSi
        USE ParnacTransformFilter, ONLY: &
            FFilterSpatial1D, FFilterSpatial2D, &
            TransformT, TransformTInv, TransformT_sml, TransformTInv_sml, &
            TransformXYZ, TransformXYZInv, TransformXYZ_sml, TransformXYZInv_sml, &
            FFilterSpaceSpatial3D, dTaperingWindow
        USE ParnacDataRedist, ONLY: &
            ReorderDistr0ToDistr1, ReorderDistr1ToDistr0, &
            ReorderDistr1ToDistr2, ReorderDistr2ToDistr1
    #endif
    
    ! *****************************************************************************
    !
    !   GLOBAL DECLARATIONS
    !
    !   none
    !
        IMPLICIT NONE
    
    ! *****************************************************************************
    !
    !   CONTAINED SUBROUTINES AND FUNCTIONS
    !
    !   InitSource            sub   Fills the Source space from an array Excitation
    !                               generated by InitSourceExcitation
    !   InitSource_Planewave  sub   Initializes the Source space for a Plane wave
    !                               source condition
    !   SourceDimensions      sub   Returns the dimensions of the source aperture
    !                               and signature in number of points
    !   InitSourceExcitation  sub   Generates the source excitation = aperture +
    !                               signature into an array Excitation used by the
    !                               InitSource and InitSource_Planewave subroutines
    !   LoadSourceExcitation  sub   Obtain the source excitation from a data file
    !   InitSourceSignature   sub   Generate the source signature for a parametric
    !                               signature description
    !   InitCylindricalFlatAperture  sub   Generate a cylindrical source aperture
    !   InitRectangularFlatAperture  sub   Generate a rectangular source aperture
    !   InitTriangularFlatAperture   sub   Generate a triangular source aperture
    !   InitPointSourceAperture      sub   Generate a point source aperture
    !   InitPhasedArrayAperture      sub   Generate a phased array source aperture
    !   InitMatrixArrayAperture      sub   Generate a matrix array source aperture
    !
    ! =============================================================================
    !
    CONTAINS
    
    #ifndef RUNESTIMATOR
    
        SUBROUTINE InitSource(cSource)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitSource fills the Source space from a 3D array generated
    !   by InitSourceExcitation. The generated source is either given by a data
    !   file or by a parametric source description in the config file.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   cSource   io   type(space)  source space to be filled - it is assumed that
    !                               it is already initialized and that a Grid 0
    !                               array is allocated
    !
            type(Space), intent(inout)::            cSource; 
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   iXInd     i8b    beam index in X
    !   iYind     i8b    beam index in Y
    !   iLi       i8b    loop counter over the positions in the beam
    !   iLt       i8b    loop counter over the time instants in the beam
    !   Ksource   i8b    size of the source in points in the T-dimension
    !   Lsource   i8b    size of the source in points in the X-dimension
    !   Msource   i8b    size of the source in points in the Y-dimension
    !   acTemp    char   temporary char array for output log messages
    !   error     i8b    error number
    !   Excitation  dp   3D array containing the source description in (t,x,y)
    !
            integer(i8b)::                          iXInd, iYInd, iLi, iLt
            integer(i8b) ::                         Ksource, LSource, MSource
            character(len=1024) ::                  acTemp; 
            integer(i8b) ::                         error
            real(dp), allocatable ::                Excitation(:, :, :)
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   log file entries and, if applicable, reading of the source data files
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   alloctest
    !   PrintToLog
    !   SourceDimensions
    !   InitSourceExcitation
    !
    ! =============================================================================
    
            write (acTemp, '("InitSource")'); call PrintToLog(acTemp, 1); 
            if (cSource.iSpaceIdentifier /= iSI_PRIMARYSOURCE) then
                call PrintToLog("Error. Cannot create a source in a non-source space", -1)
                stop; 
            end if
    
            if (cSource.cGrid.iDistr /= 0) then
                call PrintToLog("Error. Distribution of Source should be 0", -1)
                stop; 
            end if
    
            !Get and test dimensions
    
            call SourceDimensions(Ksource, Lsource, Msource, cSource%dDt, cSource%dDx)
            if ((cSource%iDimT < Ksource) .or. (cSource%iDimX < Lsource) .or. (cSource%iDimY < Msource)) then
                write (acTemp, "('Error in InitSource, source space dimensions are too small for the source')")
                call PrintToLog(acTemp, -1)
                write (acTemp, "('Source dims TXY',I6,I6,I6,', Source Space dims TXY',I6,I6,I6)") &
                    Ksource, Lsource, Msource, cSource%iDimT, cSource%iDimX, cSource%iDimY
                call PrintToLog(acTemp, -1)
                stop
            end if
    
            !Allocate the Excitation array
            allocate (Excitation(1:cSource%iDimT, 1:cSource%iDimX, 1:cSource%iDimY), STAT=error)
            call alloctest(error, 'Excitation in InitSource')
    
            ! Fill the Excitation with the source description
            call InitSourceExcitation(Excitation, cSource)
    
            ! Store Excitation in the first z-plane in cSource
            !         note that Excitation should always be centered around x=0,y=0, i.e. the zero position
            !        is found at position floor(size(Excitation,[2 or 3])/2)+1
            !        In t Excitation starts at t =0 and it is assumed below that the source is located in the plane z=0
            do iLi = 0, cSource%cGrid%iD0LocN - 1
                !only on the first plane in the grid (whatever iStartZ may be...)
                if (cSource%cGrid%aiD0Loc(1 + iLi, 3) == 0) then
    
                    !Position in the global grid, assuming that the XY slice is at Z=0.
                    iXInd = cSource%cGrid%aiD0Loc(1 + iLi, 1) + cSource%iStartX
                    iYInd = cSource%cGrid%aiD0Loc(1 + iLi, 2) + cSource%iStartY
    
                    if ((abs(iXInd) < size(Excitation, 2)/2) .AND. (abs(iYInd) < size(Excitation, 3)/2)) then
    
                        !Put the non-zero values from the source into the grid
                        !- The time dimension in Excitation may be longer than the grid allows.
                        !- We assume here that iStartT<=0.
                        !- The index in Excitation always starts at one, translate this to the beam index.
                        do iLt = 0, min(cSource%cGrid%iD0TL + cSource%iStartT, int(size(Excitation, 1), i8b)) - 1
                            cSource%cGrid%parD0(1 + iLi*cSource%cGrid%iD0IS + (iLt - cSource%iStartT)*cSource%cGrid%iD0TS) = &
                                Excitation(1 + iLt, iXInd + size(Excitation, 2)/2 + 1, iYInd + size(Excitation, 3)/2 + 1); 
                        end do
    
                    end if
                end if
            end do
    
            ! Deallocate the Excitation array
            if (allocated(Excitation)) then
                deallocate (Excitation); 
            end if
        END SUBROUTINE InitSource
    
        SUBROUTINE InitSource_Planewave(cSource)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitSource_Planewave fills the Source space for the case of
    !   the plane wave source description. We 'trick' the InitSourceExcitation
    !   by making it believe that it creates a point source. For this the function
    !   pre-assumes that the source shape type is set to iHI_POINTSOURCE. In the
    !   subroutine, the point source description is multiplied with a normalization
    !   factor, and it is stored into the Source space, which is assumed to have
    !   a size in X and Y of 1 point.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   cSource   io   type(space)  source space to be filled - it is assumed that
    !                               it is already initialized and that a Grid 0
    !                               array is allocated
    !
            type(Space), intent(inout)::                        cSource; 
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   iLt       i8b    loop counter over the time instants in the beam
    !   Ksource   i8b    size of the source in points in the T-dimension
    !   Lsource   i8b    size of the source in points in the X-dimension
    !   Msource   i8b    size of the source in points in the Y-dimension
    !   acTemp    char   temporary char array for output log messages
    !   error     i8b    error number
    !   Excitation  dp   3D array containing the source description in (t,x,y)
    !
            integer(i8b)                :: iLt
            integer(i8b)                :: Ksource, LSource, MSource
            character(len=1024)         :: acTemp
            integer(i8b)                :: error
            real(dp), allocatable       :: Excitation(:, :, :)
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   log file entries and, if applicable, reading of the source data files
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   alloctest
    !   PrintToLog
    !   SourceDimensions
    !   InitSourceExcitation
    !
    ! =============================================================================
    
            write (acTemp, '("InitSource")'); call PrintToLog(acTemp, 1); 
            if (cSource.iSpaceIdentifier /= iSI_PRIMARYSOURCE) then
                call PrintToLog("Error. Cannot create a source in a non-source space", -1)
                stop; 
            end if
    
            if (cSource.cGrid.iDistr /= 0) then
                call PrintToLog("Error. Distribution of Source should be 0", -1)
                stop; 
            end if
    
            !Get and test dimensions
            call SourceDimensions(Ksource, Lsource, Msource, cSource%dDt, cSource%dDx)
            if ((cSource%iDimT < Ksource) .or. (cSource%iDimX < Lsource) .or. (cSource%iDimY < Msource)) then
                write (acTemp, "('Error in InitSource_Planewave, source space dimensions are too small for the source')")
                call PrintToLog(acTemp, -1)
                write (acTemp, "('Source dims TXY',I6,I6,I6,', Source Space dims TXY',I6,I6,I6)") &
                    Ksource, Lsource, Msource, cSource%iDimT, cSource%iDimX, cSource%iDimY
                call PrintToLog(acTemp, -1)
                stop
            end if
    
            allocate (Excitation(1:cSource%iDimT, 1:cSource%iDimX, 1:cSource%iDimY), STAT=error)
            call alloctest(error, 'Excitation in InitSource_Planewave')
    
            ! This functions allocates and fills Excitation
            call InitSourceExcitation(Excitation, cSource)
    
            ! We had made InitSourceExcitation think that it created a point source
            ! Now we correct for the extra factor included due to the point source description
            Excitation = Excitation/((cModelParams%freq0/cMediumParams%c0)*(1.0_dp/cSource.dDx)**2)
    
            ! Store Excitation in the first position in cSource on each processor
            !         note that Excitation should always be centered around x=0,y=0. It starts at t =0 and z=0 for all values
            do iLt = 0, min(cSource%cGrid%iD0TL, int(size(Excitation, 1), i8b)) - 1
                cSource%cGrid%parD0(1 + iLt*cSource%cGrid%iD0TS) = &
                    Excitation(1 + iLt, 1, 1); 
            end do
    
            ! Deallocate the array allocated in init_source_2
            if (allocated(Excitation)) then
                deallocate (Excitation); 
            end if
    
        END SUBROUTINE InitSource_Planewave
    
    #endif
    
        SUBROUTINE SourceDimensions(iDimT, iDimX, iDimY, dDt, dDx)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine SourceDimensions returns the sizes in number of points for
    !   each of the source signature and aperture types in the three dimensions
    !   t,x and y. The data are extracted from the cSourceParams structure.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   iDimT   o  i8b  Size of the T-dimension
    !   iDimX   o  i8b  Size of the X-dimension
    !   iDimY   o  i8b  Size of the Y-dimension
    !   dDt     i  dp   Stepsize in T
    !   dDx     i  dp   Stepsize in X
    !
            real(dp), intent(in) ::         dDt, dDx
            integer(i8b), intent(out) ::    iDimT, iDimX, iDimY
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   i        i8b   temporary variable
    !   dims     i8b   size of the dimensions
    !   dTdelay  dp    maximal time delay in the data file source description
    !   acTemp   char  temporary char array for output log messages
    !
            character(LEN=2048) ::                  acTemp; 
            integer(i8b) ::                         i, dims(3)
            real(dp) ::                             dTdelay
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   if applicable, checking the dimensions of the source in the source
    !   data file.
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   BINARY_read3Darraysize
    !   BINARY_read3DarrayTdelay
    !   ASCII_read3Darraysize
    !   BINARY_read1Darraysize
    !   ASCII_read1Darraysize
    !
    ! =============================================================================
    
            if (trim(cSourceParams%srcfilename) /= 'none') then
                ! If we have a data file source description, then check the sizes from
                ! data file
    !    print *, "CCCCCCCCCCCCCCCCCCCCCCCCCCCC"
    #ifndef RUNESTIMATOR
                i = index(cSourceParams%srcfilename, '.')
                if (cSourceParams%srcfilename(i + 1:i + 3) == 'bin') then
                    dims = BINARY_read3Darraysize(trim(sInputDir)//trim(cSourceParams%srcfilename))
                    dTdelay = BINARY_read3DarrayTdelay(trim(sInputDir)//trim(cSourceParams%srcfilename))
                    !dTdelay is given in milliseconds, normalize
                    dTdelay = dTdelay*cModelParams.freq0*1.0e-6
                else
                    dims = ASCII_read3Darraysize(trim(sInputDir)//trim(cSourceParams%srcfilename))
                end if
    
                iDimT = dims(1) + ceiling(dTdelay/dDt)
                iDimX = dims(2)
                iDimY = dims(3)
    !                print *, "AAAAAAAAAAAAAAAAAAAAAAAA"
    #else
    !                iDimT = 0
    !                iDimX = 0
    !                iDimY = 0
    !                print *, "BBBBBBBBBBBBBBBBBBBBBBBB"
    #endif
    
            else
                ! For parametric source descriptions, there are specific rules for the
                ! source sizes. Each time, we also include an extra margin for possible
                ! filtering effects
    
                if (cSourceParams%srcsigntype == iGI_GAUSSIAN) then
                    iDimT = ceiling((cSourceParams%Tpulse + cSourceParams%maxphasedelay)/dDt)
                else if (cSourceParams%srcsigntype == iGI_BLACKMAN) then
                    iDimT = ceiling((1.1_dp*cSourceParams%Tpulse + cSourceParams%maxphasedelay)/dDt)
                else
    #ifndef RUNESTIMATOR
                    i = index(cSourceParams%srcsignfilename, '.')
                    if (cSourceParams%srcsignfilename(i + 1:i + 3) == 'bin') then
                        iDimT = BINARY_read1Darraysize(trim(sInputDir)//trim(cSourceParams%srcsignfilename)) + ceiling(real(cSourceParams%maxphasedelay)/dDt)
                    else
                        iDimT = ASCII_read1Darraysize(trim(sInputDir)//trim(cSourceParams%srcsignfilename)) + ceiling(real(cSourceParams%maxphasedelay)/dDt)
                    end if
    #else
    !                iDimT = 0
    #endif
                end if
    
                if (cSourceParams%srcshapetype == iHI_CYLINDRICAL) then
                    ! take 10 points extra in each dimension to allow for Gibb's
                    iDimX = 2*((cSourceParams%radius/dDx) + 10) + 1
                    iDimY = 2*((cSourceParams%radius/dDx) + 10) + 1
                else if (cSourceParams%srcshapetype == iHI_RECTANGULAR) then
                    ! take 10 points extra in each dimension to allow for Gibb's
                    iDimX = 2*(cSourceParams%rectwidth/(2.0_dp*dDx) + 10) + 1
                    iDimY = 2*(cSourceParams%rectheight/(2.0_dp*dDx) + 10) + 1
                else if (cSourceParams%srcshapetype == iHI_TRIANGULAR) then
                    ! take 10 points extra in each dimension to allow for Gibb's
                    iDimX = 2*(sqrt(3.0_dp)*cSourceParams%triangedge/(4.0_dp*dDx) + abs(cSourceParams%triangxcenter)/dDx + 10) + 1
                    iDimY = 2*(cSourceParams%triangedge/(2.0_dp*dDx) + 10) + 1
                else if (cSourceParams%srcshapetype == iHI_POINTSOURCE) then
                    iDimX = 1
                    iDimY = 1
                else if (cSourceParams%srcshapetype == iHI_PHASEDARRAY) then
                    ! take 15 points extra in each dimension to allow for Gibb's
                    iDimX = 2*(ceiling(cSourceParams%arraywidth/(2.0_dp*dDx)) + 15) + 1
                    iDimY = 2*(ceiling(cSourceParams%elheight/(2.0_dp*dDx)) + 15) + 1
    !KH                else if (cSourceParams%srcshapetype==iHI_MATRIXARRAY) then
    !KH                        ! take 15 points extra in each dimension to allow for Gibb's
    !KH                        iDimX=2*(ceiling((cSourceParams%arraywidth)/(2.0_dp*dDx))+15)+1
    !KH                        iDimY=2*(ceiling(cSourceParams%elheight/(2.0_dp*dDx))+15)+1
                end if
    
            end if
    
        END SUBROUTINE SourceDimensions
    
    #ifndef RUNESTIMATOR
    
        SUBROUTINE InitSourceExcitation(Excitation, cSpace)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitSourceExcitation generates the source from the
    !   parameters in the cSourceParams structure - either a data file description
    !   or a parametric source description. The source is stored in a 3D array.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   Excitation  o  dp            3D array to be filled with the source
    !   cSpace      i  type(Space)   Space for which the source is generated - is
    !                                mainly used to determine step sizes and filter
    !                                frequency
    !
            real(dp), intent(out) ::                    Excitation(:, :, :)
            type(Space), intent(in) ::                  cSpace
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   i            i8b   temporary variable / loop counter
    !   j            i8b   temporary variable / loop counter
    !   k            i8b   temporary variable / loop counter
    !   Ksource      i8b   size of the source in the T-dimension
    !   Lsource      i8b   size of the source in the X-dimension
    !   Msource      i8b   size of the source in the Y-dimension
    !   plan_transform         i8b  FFTW transform plan
    !   plan_1D_transform      i8b  FFTW transform plan for 1D transform
    !   plan_1D_transform_inv  i8b  FFTW transform plan for 1D invers transform
    !   dKcutoff     dp    cutoff spatial angular frequency
    !   dwsource     dp    step size in the w-axis
    !   waxis        dp    angular frequency axis
    !   Ptempdelay   dp    1D array used to include the d/dt in the data
    !                      file source description for the Q-source
    !   Signature    dp    1D array for the time signature of the source
    !   fPtempdelay  dpc   1D array used to include the d/dt in the data
    !                      file source description for the Q-source
    !   fExcitation  dpc   3D array for the signature+aperture of the source
    !   fSignature   dpc   1D array for the time signature of the source
    !   error        i8b   error number
    !   acTemp       char  temporary char array for output log messages
    !
            integer(i8b)                    :: i, j, k, Ksource, Lsource, Msource, error
            integer(i8b)                    :: plan_transform, plan_1D_transform, plan_1D_transform_inv
            real(dp)                        :: dKcutoff, dwsource
            real(dp), allocatable           :: waxis(:), Ptempdelay(:), Signature(:)
            complex(dpc), allocatable       :: fPtempdelay(:), fExcitation(:, :, :), fSignature(:)
            character(LEN=2048)             :: acTemp; 
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries and, if applicable, reading of the source data files
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   LoadSourceExcitation
    !   dfftw_plan_dft_r2c_1d
    !   dfftw_plan_dft_c2r_1d
    !   dfftw_execute
    !   dfftw_destroy_plan
    !   InitSourceSignature
    !   InitCylindricalFlatAperture
    !   InitRectangularFlatAperture
    !   InitTriangularFlatAperture
    !   InitPointSourceAperture
    !   InitPhasedArrayAperture
    !   InitMatrixArrayAperture
    !   dfftw_plan_many_dft_c2r
    !   alloctest
    !   PrintToLog
    !
    ! *****************************************************************************
    !
    !   PSEUDOCODE
    !
    !   If the source definition type is a datafile then
    !       Obtain source excitation from the datafile
    !       If the source excitation type is a volume source (a Q-source)
    !           Apply 2 * d/dt to the Q-source to obtain the S-source
    !       Elseif the source excitation type is a force source (an F-source)
    !           Multiply the source with -2. The differentiation in z is applied
    !             after the convolution with the Green's function has been applied.
    !       Elseif the source excitation type is a time differentiated Q-source
    !           Multiply with 2 and with 1/freq0 due to normalization of the
    !             t-dimension.
    !   Else -> the source definition type is parametric
    !       Obtain the source signature from the parametric description
    !       Transform the source signature to the temporal Fourier domain
    !       If the source excitation type is a volume source (a Q-source)
    !           Apply 2 * d/dt to the signature to translate the Q-source to the
    !             S-source.
    !       Elseif the source excitation type is a force source (an F-source)
    !           Multiply the signature with -2. The differentiation in z is applied
    !             after the convolution with the Green's function has been applied.
    !       Elseif the source excitation type is a time differentiated Q-source
    !           Multiply with 2 and with 1/freq0 due to normalization of the
    !             t-dimension.
    !       Obtain the aperture in for each of the various parametric aperture
    !         descriptions. This is done in the (x,y,omega)-domain to enable the
    !         inclusion of spatially-dependent time delays.
    !       Multiply the aperture with the signature to obtain the source
    !         excitation.
    !       Inverse transform the source excitation to obtain the (x,y,t) form.
    !
    ! =============================================================================
    
            write (acTemp, '("InitSourceExcitation")'); call PrintToLog(acTemp, 2); 
            Ksource = size(Excitation, 1)
            Lsource = size(Excitation, 2)
            Msource = size(Excitation, 3)
    
            if (trim(cSourceParams%srcfilename) /= 'none') then
                ! Load source file from disk
    
                call LoadSourceExcitation(Excitation, cSpace%dDt)
    
                !For the Q-source type, apply d/dt and factor 2
                !For the F-source type, apply factor -2; d/dz is applied after the convolution
                !For the FieldII or the measured signature source type, correct for the d/dt -> d/d (t prime) normalization
                !                and apply factor 2
                if (cModelParams.PrimarySourceType == iEI_QSOURCE) then
    
                    ! Apply d/dt in the temporal fourier domain by multiplication with i*omega
                    ! this method is called spectral differentiation
                    allocate (Ptempdelay(1:Ksource), fPtempdelay(1:Ksource/2 + 1), waxis(1:Ksource/2 + 1))
    
                    dwsource = two_pi/(real(Ksource, dp)*cSpace%dDt)
                    waxis = real((/(i, i=0, Ksource/2)/), dp)*dwsource
                    call dfftw_plan_dft_r2c_1d(plan_1D_transform, int(Ksource), Ptempdelay, fPtempdelay, FFTW_ESTIMATE)
                    call dfftw_plan_dft_c2r_1d(plan_1D_transform_inv, int(Ksource), fPtempdelay, Ptempdelay, FFTW_ESTIMATE)
                    do i = 1, Lsource
                        do j = 1, Msource
    
                            Ptempdelay = Excitation(:, i, j)
                            call dfftw_execute(plan_1D_transform)
                            fPtempdelay = 2.0_dp*im*waxis*fPtempdelay
                            call dfftw_execute(plan_1D_transform_inv)
                            Excitation(:, i, j) = Ptempdelay/Ksource
    
                        end do
                    end do
    
                    call dfftw_destroy_plan(plan_1D_transform)
                    call dfftw_destroy_plan(plan_1D_transform_inv)
                    deallocate (Ptempdelay, fPtempdelay, waxis)
    
                elseif (cModelParams.PrimarySourceType == iEI_FSOURCE) then
    
                    Excitation = -2.0_dp*Excitation
    
                elseif (cModelParams.PrimarySourceType == iEI_DTQSOURCE) then
    
                    Excitation = 2.0_dp*Excitation/cModelParams%freq0
    
                end if
    
            else
                !Build profile from parametric source description
    
                !allocate temporary arrays
                allocate (fExcitation(1:Ksource/2 + 1, 1:Lsource, 1:Msource), &
                          Signature(1:Ksource), fSignature(1:Ksource/2 + 1), &
                          waxis(1:Ksource/2 + 1), STAT=error)
                call alloctest(error, 'fExcitation, Signature in InitSourceExcitation')
    
                !Set cutoff frequency for spatial filter - slightly larger than the max spatial frequency
                dKcutoff = two_pi*(1.0_dp + 0.5_dp/cSpace%iDimX)*cSpace%dFnyq
    
                !Step size and axis in w
                dwsource = two_pi/(real(Ksource, dp)*cSpace%dDt)
                waxis = real((/(i, i=0, Ksource/2)/), dp)*dwsource
    
                !First: source signature
                call InitSourceSignature(Signature, cSpace%dDt)
    
                !Transform time signature to w-domain and deallocate t-domain version
                call dfftw_plan_dft_r2c_1d(plan_transform, int(Ksource), Signature, fSignature, FFTW_ESTIMATE)
                call dfftw_execute(plan_transform)
                deallocate (Signature)
    
                ! If Ksource is even, set the highest coefficient to zero
                ! From the operations we apply to it, it could obtain an
                ! imaginary component. This would result in imbalance
                ! in the FFT coefficients and yield a non-real inverse
                ! Fourier transform
                if (mod(Ksource, 2_i8b) == 0) then
                    fSignature(Ksource/2 + 1) = 0.0_dp
                end if
    
                !For the Q-source type, apply d/dt and factor 2
                !For the F-source type, apply factor -2; d/dz is applied after the convolution
                !For the FieldII or the measured signature source type, correct for the d/dt -> d/d (t prime) normalization
                !                and apply factor 2
                !Do all of this only if the source type is not a point source
                if (cSourceParams%srcshapetype /= iHI_POINTSOURCE) then
                    if (cModelParams.PrimarySourceType == iEI_QSOURCE) then
                        fSignature = 2.0_dp*im*waxis*fSignature
                    elseif (cModelParams.PrimarySourceType == iEI_FSOURCE) then
                        fSignature = -2.0_dp*fSignature
                    elseif (cModelParams.PrimarySourceType == iEI_DTQSOURCE) then
                        fSignature = 2.0_dp*fSignature/cModelParams%freq0
                    end if
                end if
    
                !Second: build source shape, both amplitude and phase delay; include spatial filter
                if (cSourceParams%srcshapetype == iHI_CYLINDRICAL) then
    
                    !cylindrical aperture: focused, defocused or unfocused
                    call InitCylindricalFlatAperture(fExcitation, cSpace%dDx, dwsource, dKcutoff)
    
                else if (cSourceParams%srcshapetype == iHI_RECTANGULAR) then
    
                    !rectangular aperture
                    call InitRectangularFlatAperture(fExcitation, cSpace%dDx, dKcutoff)
    
                else if (cSourceParams%srcshapetype == iHI_TRIANGULAR) then
    
                    !triangular aperture
                    call InitTriangularFlatAperture(fExcitation, cSpace%dDx, dKcutoff)
    
                elseif (cSourceParams%srcshapetype == iHI_POINTSOURCE) then
    
                    !point source at z=0; Excitation is actually Ssource here...
                    call InitPointSourceAperture(fExcitation, cSpace%dDx)
    
                elseif (cSourceParams%srcshapetype == iHI_PHASEDARRAY) then
    
                    !Phased array aperture: steered, focused, elevation focusing, apodized
                    call InitPhasedArrayAperture(fExcitation, cSpace%dDx, dwsource, dKcutoff)
    
    !KH                elseif (cSourceParams%srcshapetype==iHI_MATRIXARRAY) then
    
    !KH                        !Matrixarray aperture: steered, focused, apodized
    !KH                        call InitMatrixArrayAperture(fExcitation, cSpace%dDx, dwsource, dKcutoff)
    
                end if
    
                !Third: multiply Aperture with Signature
                do k = 1, Ksource/2 + 1
                    fExcitation(k, :, :) = fExcitation(k, :, :)*fSignature(k)
                end do
    
                !Fourth: Inverse transform Excitation to (x,t)-domain
                call dfftw_plan_many_dft_c2r(plan_transform, 1, int(Ksource), int(Lsource*Msource), &
                                             fExcitation, 0, 1, int(Ksource/2 + 1), &
                                             Excitation, 0, 1, int(Ksource), FFTW_ESTIMATE)
                call dfftw_execute(plan_transform)
                Excitation = Excitation/Ksource
    
                !Deallocate temp variables
                deallocate (fExcitation, fSignature, waxis)
                call dfftw_destroy_plan(plan_transform)
    
            end if
    
        END SUBROUTINE InitSourceExcitation
    
        SUBROUTINE LoadSourceExcitation(Excitation, dDt)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine LoadSourceExcitation loads the data file from disk and puts
    !   the source in the Excitation array. For the binary datafile type, the time
    !   delays are also imposed.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   Excitation  o  dp   3D array to be filled with the source
    !   dDt         i  dp   Stepsize in t
    !
            real(dp), intent(out) ::    Excitation(:, :, :)
            real(dp), intent(in) ::     dDt
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   Ksource      i8b   size of the source in the T-dimension
    !   Lsource      i8b   size of the source in the X-dimension
    !   Msource      i8b   size of the source in the Y-dimension
    !   Ktsource     i8b   size of the source in the T-dimension in the file
    !   Ltsource     i8b   size of the source in the X-dimension in the file
    !   Mtsource     i8b   size of the source in the Y-dimension in the file
    !   Lstart       i8b   start of the source in the X-dimension in the array
    !   Mstart       i8b   start of the source in the Y-dimension in the array
    !   i            i8b   temporary variable / loop counter
    !   j            i8b   temporary variable / loop counter
    !   dims         i8b   dimensions of the source stored in the file
    !   plan_1D_transform      i8b  FFTW transform plan for 1D transform
    !   plan_1D_transform_inv  i8b  FFTW transform plan for 1D invers transform
    !   dwsource     dp    step size in the w-axis
    !   waxis        dp    angular frequency axis
    !   Ptempsource  dp    temporary storage for source
    !   dTdelay      dp    2D array for the spatially-dependent time delays
    !   Ptempdelay   dp    1D array used to include the time delay
    !   fPtempdelay  dpc   1D array used to include the time delay
    !   error        i8b   error number
    !   acTemp       char  temporary char array for output log messages
    !
            integer(i8b)                :: Ksource, Lsource, Msource, Ktsource, Ltsource, Mtsource, Lstart, Mstart
            integer(i8b)                :: i, j, dims(3), error
            integer(i8b)                :: plan_1D_transform, plan_1D_transform_inv
            real(dp)                    :: dwsource
            real(dp), allocatable       :: Ptempsource(:, :, :), dTdelay(:, :), Ptempdelay(:), waxis(:)
            complex(dpc), allocatable   :: fPtempdelay(:)
            character(LEN=2048)         :: acTemp; 
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries and reading of the source excitation files
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   BINARY_read3Darraysize
    !   BINARY_read3Darray
    !   ASCII_read3Darraysize
    !   ASCII_read3Darray
    !   dfftw_plan_dft_r2c_1d
    !   dfftw_plan_dft_c2r_1d
    !   dfftw_execute
    !   dfftw_destroy_plan
    !   alloctest
    !   PrintToLog
    !
    ! =============================================================================
    
            write (acTemp, '("LoadSourceExcitation")'); call PrintToLog(acTemp, 3); 
            Ksource = size(Excitation, 1)
            Lsource = size(Excitation, 2)
            Msource = size(Excitation, 3)
    
            i = index(cSourceParams%srcfilename, '.')
            if (cSourceParams%srcfilename(i + 1:i + 3) == 'bin') then
                ! If source data file is a binary, then read binary
                ! binary files may include the spatially dependent time delays
    
                dims = BINARY_read3Darraysize(trim(sInputDir)//trim(cSourceParams%srcfilename))
                Ktsource = dims(1)
                Ltsource = dims(2)
                Mtsource = dims(3)
    
                allocate (Ptempsource(1:Ktsource, 1:Ltsource, 1:Mtsource), dTdelay(1:Ltsource, 1:Mtsource), STAT=error)
                call alloctest(error, 'Ptempsource in init_source')
                call BINARY_read_sourcedef(trim(sInputDir)//trim(cSourceParams%srcfilename), Ptempsource, dTdelay)
    
                !normalize dTdelay.. in the file given in milliseconds
                dTdelay = dTdelay*cModelParams.freq0*1.0D-6
    
            else
                ! Otherwise, source data file is an ASCII
                ! ASCII files don't include any spatially dependent time delays
    
                dims = ASCII_read3Darraysize(trim(sInputDir)//trim(cSourceParams%srcfilename))
                Ktsource = dims(1)
                Ltsource = dims(2)
                Mtsource = dims(3)
    
                allocate (Ptempsource(1:Ktsource, 1:Ltsource, 1:Mtsource), STAT=error)
                call alloctest(error, 'Ptempsource in init_source')
                call ASCII_read3Darray(trim(sInputDir)//trim(cSourceParams%srcfilename), Ptempsource)
    
            end if
    
            ! Place the source file at the center of Excitation
            Lstart = ceiling((Lsource + 1)/2.0_dp) - ceiling((Ltsource + 1)/2.0_dp)
            Mstart = ceiling((Msource + 1)/2.0_dp) - ceiling((Mtsource + 1)/2.0_dp)
    
            Excitation = 0.0_dp
            Excitation(1:Ktsource, Lstart + 1:Lstart + Ltsource, Mstart + 1:Mstart + Mtsource) = Ptempsource
            deallocate (Ptempsource)
    
            ! Binary file: apply the time delays to Excitation, do this in the spectral domain
            i = index(cSourceParams%srcfilename, '.')
            if (cSourceParams%srcfilename(i + 1:i + 3) == 'bin') then
    
                allocate (Ptempdelay(1:Ksource), fPtempdelay(1:Ksource/2 + 1), waxis(1:Ksource/2 + 1))
    
                dwsource = two_pi/(real(Ksource, dp)*dDt)
                waxis = real((/(i, i=0, Ksource/2)/), dp)*dwsource
                call dfftw_plan_dft_r2c_1d(plan_1D_transform, int(Ksource), Ptempdelay, fPtempdelay, FFTW_ESTIMATE)
                call dfftw_plan_dft_c2r_1d(plan_1D_transform_inv, int(Ksource), fPtempdelay, Ptempdelay, FFTW_ESTIMATE)
    
                do i = 1, Ltsource
                    do j = 1, Mtsource
    
                        Ptempdelay = Excitation(:, Lstart + i, Mstart + j)
    
                        call dfftw_execute(plan_1D_transform)
    
                        fPtempdelay = fPtempdelay*exp(-im*waxis*dTdelay(i, j))
    
                        call dfftw_execute(plan_1D_transform_inv)
    
                        !Normalization of FFT
                        Excitation(:, Lstart + i, Mstart + j) = Ptempdelay/Ksource
    
                    end do
                end do
    
                call dfftw_destroy_plan(plan_1D_transform)
                call dfftw_destroy_plan(plan_1D_transform_inv)
                deallocate (Ptempdelay, fPtempdelay, dTdelay, waxis)
            end if
    
        END SUBROUTINE LoadSourceExcitation
    
        SUBROUTINE InitSourceSignature(Signature, dDt)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitSourceSignature generates the signature for the
    !   parametric signature descriptions. Currently, two types are implemented:
    !   a harmonic signal with a gaussian envelope, and a blackman pulsed signal.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   Signature  o  dp   1D array to be filled with the source signature
    !   dDt        i  dp   Stepsize in t
    !
            real(dp), intent(out) ::    Signature(:)
            real(dp), intent(in) ::     dDt
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   i              i8b   temporary variable / loop counter
    !   Ksource        i8b   size of the signature
    !   Ktsource       i8b   size of the signature in signature file
    !   taxis          dp    time axis
    !   Signaturetemp  dp    temporary storage for the signature
    !   b0,b1,b2,b3    dp    constants for the blackman pulse
    !   acTemp         char  temporary char array for output log messages
    !
            integer(i8b)            :: i, Ksource, Ktsource, IS,IE
            real(dp), allocatable   :: taxis(:), Signaturetemp(:)
            character(LEN=2048)     :: acTemp; 
            !blackman pulse, constants
            real(dp), parameter     :: b0 = 0.35869d0
            real(dp), parameter     :: b1 = -0.48829d0
            real(dp), parameter     :: b2 = 0.14128d0
            real(dp), parameter     :: b3 = -0.01168d0
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries and, if applicable, reading of the source signature file
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   BINARY_read1Darraysize
    !   BINARY_read1Darray
    !   ASCII_read1Darraysize
    !   ASCII_read1Darray
    !   dfftw_plan_dft_r2c_1d
    !   dfftw_plan_dft_c2r_1d
    !   dfftw_execute
    !   dfftw_destroy_plan
    !   alloctest
    !   PrintToLog
    !
    ! =============================================================================
    
            write (acTemp, '("InitSourceSignature")'); call PrintToLog(acTemp, 3); 
            Ksource = size(Signature, 1)
            Signature = 0.0_dp
    
            allocate (taxis(1:Ksource))
            taxis = real((/(i, i=0, Ksource - 1)/), dp)*dDt
    
            if (cSourceParams%srcsigntype == iGI_GAUSSIAN) then
                ! Generate a harmonic signal with a gaussian envelope
    
                Signature = cSourceParams.Pstart*exp(-((taxis(1:Ksource) - cSourceParams%Tdelay)/ &
                                                       (cSourceParams%Twidth/2.0_dp))**cSourceParams%power) &
                            *sin(two_pi*(taxis(1:Ksource) - cSourceParams%Tdelay)) &
                            *(1.0_dp + dsign(1.0_dp, taxis(1:Ksource)))/2.0_dp
    
            else if (cSourceParams%srcsigntype == iGI_BLACKMAN) then
                ! Generate a blackman pulse
    
                Signature = cSourceParams.Pstart*(b0 + b1*cos(1.0_dp*two_pi*(taxis(1:Ksource) - cSourceParams%Tdelay)/cSourceParams%Tpulse) &
                                                  + b2*cos(2.0_dp*two_pi*(taxis(1:Ksource) - cSourceParams%Tdelay)/cSourceParams%Tpulse) &
                                                  + b3*sin(3.0_dp*two_pi*(taxis(1:Ksource) - cSourceParams%Tdelay)/cSourceParams%Tpulse)) &
                            *(dsign(1.0_dp, taxis(1:Ksource) - cSourceParams%Tdelay) - dsign(1.0_dp, taxis(1:Ksource) - cSourceParams%Tdelay - cSourceParams%Tpulse))/2.0_dp
                        else if (cSourceParams%srcsigntype == iGI_INTEGRABLE) then
                            Signature = 0.0D0;
                            IS = cSourceParams%Tdelay/dDt; 
                            IE = (cSourceParams%Tdelay+cSourceParams%Twidth)/dDt
                            Signature(IS:IE) = cSourceParams%Pstart*sin(two_pi*(taxis(IS:IE)))!*exp(-A_const**cSourceParams%power)
                            ! Signature = cSourceParams%Pstart*sin(two_pi*(taxis - cSourceParams%Tdelay))!*exp(-A_const**cSourceParams%power)
                            ! Signature((cSourceParams%Twidth + cSourceParams%Tdelay)/dDt:Ksource) = 0; 
                            ! Signature(1:(cSourceParams%Tdelay)/dDt) = 0; 
                
                            ! Signature((cSourceParams%Tdelay)/dDt+1 : (cSourceParams%Twidth + cSourceParams%Tdelay)/dDt) = &
                            ! Signature((cSourceParams%Tdelay)/dDt+1 : (cSourceParams%Twidth + cSourceParams%Tdelay)/dDt) * dTaperingWindow( int(cSourceParams%Twidth/dDt - 1,i8b), dDt, 2.0D0, 2.0D0)
                            
                            ! A_const= ((taxis(1:Ksource) - cSourceParams%Tdelay)/ (cSourceParams%Twidth/2.0_dp));
                            ! Signature = cSourceParams%Pstart*exp(-A_const**cSourceParams%power)*( &
                            ! - cSourceParams%power*A_const/(cSourceParams%Twidth/2.0_dp)*sin(two_pi*(taxis(1:Ksource) - cSourceParams%Tdelay)) + &
                            !   two_pi*cos(two_pi*(taxis(1:Ksource) - cSourceParams%Tdelay))) &
                            ! *(1.0_dp + dsign(1.0_dp, taxis(1:Ksource)))/2.0_dp
                            
                            ! Signature(cSourceParams%Tdelay/dDt+1)=0.0D0;
            else if (cSourceParams%srcsigntype == iGI_FILE) then
                !load stored source signature file
    
                i = index(cSourceParams%srcfilename, '.')
                if (cSourceParams%srcfilename(i + 1:i + 3) == 'bin') then
                    Ktsource = BINARY_read1Darraysize(trim(sInputDir)//trim(cSourceParams%srcsignfilename))
                    allocate (Signaturetemp(1:Ktsource))
                    call BINARY_read1Darray(trim(sInputDir)//trim(cSourceParams%srcsignfilename), Signaturetemp)
                else
                    Ktsource = ASCII_read1Darraysize(trim(sInputDir)//trim(cSourceParams%srcsignfilename))
                    allocate (Signaturetemp(1:Ktsource))
                    call ASCII_read1Darray(trim(sInputDir)//trim(cSourceParams%srcsignfilename), Signaturetemp)
                end if
    
                !put into Signature
                Signature(1:Ktsource) = Signaturetemp
    
                deallocate (Signaturetemp)
            end if
    
            deallocate (taxis)
    
        END SUBROUTINE InitSourceSignature
    
        SUBROUTINE InitCylindricalFlatAperture(fAperture, dDx, dDw, dKcutoff)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitCylindricalFlatAperture generates the aperture for the
    !   parametric description of a cylindrical source. The aperture is centered
    !   around the origin and has a certain radius and focus (may be positive or
    !   negative). The aperture is spatially filtered in X and Y.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   fAperture  o  dpc  3D array to be filled with the source aperture in the
    !                      (x,omega) domain.
    !   dDx        i  dp   Stepsize in x
    !   dDw        i  dp   Stepsize in w
    !   dKcutoff   i  dp   Cutoff spatial angular frequency
    !
            complex(dpc), intent(out) ::                    fAperture(:, :, :)
            real(dp), intent(in) ::                         dDx
            real(dp), intent(in) ::                         dDw
            real(dp), intent(in) ::                         dKcutoff
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   i               i8b   temporary variable / loop counter
    !   j               i8b   temporary variable / loop counter
    !   k               i8b   temporary variable / loop counter
    !   ispan           i8b   span of the source in x
    !   jspan           i8b   span of the source in y
    !   Ksource         i8b   size of the aperture in omega
    !   Lsource         i8b   size of the aperture in x
    !   Msource         i8b   size of the aperture in y
    !   phasedelay      dp    phase delay due to focusing
    !   waxis           dp    angular frequency axis
    !   xaxis           dp    x axis
    !   yaxis           dp    y axis
    !   xaxislrg        dp    x axis in the oversampling mode
    !   yaxislrg        dp    y axis in the oversampling mode
    !   sourceshapelrg  dpc   3D array of the source aperture in oversampling mode
    !   sourceshapesml  dpc   3D array containing the source aperture
    !   oversampling    dp    oversampling factor
    !   ioversampling   i8b   oversampling factor in int(i8b) form
    !   error           i8b   error number
    !   acTemp          char  temporary char array for output log messages
    !
            integer(i8b)                :: i, j, k, ispan, jspan, Ksource, Lsource, Msource, error
            real(dp)                    :: phasedelay
            real(dp), allocatable       :: waxis(:), xaxis(:), yaxis(:), xaxislrg(:), yaxislrg(:)
            complex(dpc), allocatable   :: sourceshapelrg(:, :), sourceshapesml(:, :)
            character(LEN=2048)         :: acTemp; 
            real(dp), parameter         :: oversampling = 8.0_dp
            integer(i8b)                :: ioversampling = oversampling
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   FFilterSpatial2D
    !   alloctest
    !   PrintToLog
    !
    ! =============================================================================
    
            write (acTemp, '("InitCylindricalFlatAperture")'); call PrintToLog(acTemp, 3); 
            Ksource = size(fAperture, 1)
            Lsource = size(fAperture, 2)
            Msource = size(fAperture, 3)
            fAperture = 0.0_dp
    
            allocate (sourceshapelrg(1:ioversampling*Lsource, 1:ioversampling*Msource), &
                      sourceshapesml(1:Lsource, 1:Msource), &
                      waxis(1:Ksource), xaxis(1:Lsource), yaxis(1:Msource), &
                      xaxislrg(1:ioversampling*Lsource), yaxislrg(1:ioversampling*Msource), STAT=error)
            call alloctest(error, 'Sourceshape and help variables in InitCylindricalFlatAperture')
    
            !Axes in w and normal size and large size in x and y
            waxis = real((/(i, i=0, Ksource - 1)/), dp)*dDw
            xaxis = real((/(i, i=-Lsource/2, Lsource/2 - 1)/), dp)*dDx
            yaxis = real((/(i, i=-Msource/2, Msource/2 - 1)/), dp)*dDx
            xaxislrg = real((/(i, i=-ioversampling/2*Lsource, ioversampling/2*Lsource - 1)/), dp)*dDx/oversampling
            yaxislrg = real((/(i, i=-ioversampling/2*Msource, ioversampling/2*Msource - 1)/), dp)*dDx/oversampling
    
            ! Fill source array. Amplitude:
            ! 0.0 if outside source aperture
            ! 1.0 if on source aperture
            ! 0.5 if on the edge of the aperture
            if (abs(cSourceParams%radfocus) > 1e-5) then
                ! If a radfocus is given, use phase delays to set a focus
    
                do k = 1, Ksource
                    sourceshapelrg = 0.0_dp
                    ispan = floor(cSourceParams%radius/(dDx/oversampling))
    
                    do i = ioversampling*Lsource/2 + 1 - ispan, ioversampling*Lsource/2 + 1 + ispan
                        jspan = floor(sqrt(abs(cSourceParams%radius**2 - xaxislrg(i)**2))/(dDx/oversampling))
    
                        do j = ioversampling*Msource/2 + 1 - jspan, ioversampling*Msource/2 + 1 + jspan
                            if (cSourceParams%radfocus > 0) then
                                ! Normal focus, i.e. generate focused piston
                                !maximum phasedelay at r=0, minimum at r=radius
                                phasedelay = sqrt(cSourceParams%radius**2 + cSourceParams%radfocus**2) - &
                                             sqrt(xaxislrg(i)**2 + yaxislrg(j)**2 + cSourceParams%radfocus**2)
                            else
                                ! Negative focus, i.e. generate defocused piston
                                !minimum phasedelay at r=0, maximum at r=radius
                                phasedelay = sqrt(xaxislrg(i)**2 + yaxislrg(j)**2 + cSourceParams%radfocus**2) - &
                                             abs(cSourceParams%radfocus)
                            end if
    
                            sourceshapelrg(i, j) = exp(-im*waxis(k)*phasedelay)
                        end do
                        if ((sqrt(xaxislrg(i)**2 + yaxislrg(ioversampling/2*Msource + 1 + jspan)**2) - cSourceParams%radius) > -1e-6) then
                            sourceshapelrg(i, ioversampling/2*Msource + 1 + jspan) = 0.5_dp*exp(-im*waxis(k)*phasedelay)
                            sourceshapelrg(i, ioversampling/2*Msource + 1 - jspan) = 0.5_dp*exp(-im*waxis(k)*phasedelay)
                        end if
                    end do
    
                    !Filter the source spatially
                    call FFilterSpatial2D(sourceshapelrg, sourceshapesml, dDx/oversampling, ioversampling, dKcutoff)
    
                    fAperture(k, :, :) = sourceshapesml
                end do
            else
                ! Unfocused cylindrical piston
    
                sourceshapelrg = 0.0_dp
                ispan = floor(cSourceParams%radius/(dDx/oversampling))
                do i = ioversampling*Lsource/2 + 1 - ispan, ioversampling*Lsource/2 + 1 + ispan
                    jspan = floor(sqrt(abs(cSourceParams%radius**2 - xaxislrg(i)**2))/(dDx/oversampling))
                    do j = ioversampling/2*Msource + 1 - jspan, ioversampling/2*Msource + 1 + jspan
                        sourceshapelrg(i, j) = 1.0_dp
                    end do
                    if ((sqrt(xaxislrg(i)**2 + yaxislrg(ioversampling/2*Msource + 1 + jspan)**2) - cSourceParams%radius) > -1e-6) then
                        sourceshapelrg(i, ioversampling/2*Msource + 1 + jspan) = 0.5_dp
                        sourceshapelrg(i, ioversampling/2*Msource + 1 - jspan) = 0.5_dp
                    end if
                end do
    
                !Filter the source spatially
                call FFilterSpatial2D(sourceshapelrg, sourceshapesml, dDx/oversampling, ioversampling, dKcutoff)
    
                ! Aperture is independent of omega
                do k = 1, Ksource
                    fAperture(k, :, :) = sourceshapesml
                end do
            end if
    
            deallocate (sourceshapelrg, sourceshapesml, waxis, xaxis, yaxis, xaxislrg, yaxislrg)
    
        END SUBROUTINE InitCylindricalFlatAperture
    
        SUBROUTINE InitRectangularFlatAperture(fAperture, dDx, dKcutoff)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitRectangularFlatAperture generates the aperture for the
    !   parametric description of a rectangular source. The parameter rectwidth
    !   gives the size in x and rectheight gives the size in y. The aperture is
    !   spatially filtered in X and Y.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   fAperture  o  dpc  3D array to be filled with the source aperture in the
    !                      (x,omega) domain.
    !   dDx        i  dp   Stepsize in x
    !   dKcutoff   i  dp   Cutoff spatial angular frequency
    !
            complex(dpc), intent(out) ::                    fAperture(:, :, :)
            real(dp), intent(in) ::                         dDx
            real(dp), intent(in) ::                         dKcutoff
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   i               i8b   temporary variable / loop counter
    !   j               i8b   temporary variable / loop counter
    !   k               i8b   temporary variable / loop counter
    !   Ksource         i8b   size of the aperture in omega
    !   Lsource         i8b   size of the aperture in x
    !   Msource         i8b   size of the aperture in y
    !   Lstart          i8b   start of the source in the X-dimension
    !   Lend            i8b   end of the source in the X-dimension
    !   Mstart          i8b   start of the source in the Y-dimension in the array
    !   Mend            i8b   end of the source in the Y-dimension
    !   xaxis           dp    x axis
    !   yaxis           dp    y axis
    !   xaxislrg        dp    x axis in the oversampling mode
    !   yaxislrg        dp    y axis in the oversampling mode
    !   sourceshapelrg  dpc   3D array of the source aperture in oversampling mode
    !   sourceshapesml  dpc   3D array containing the source aperture
    !   oversampling    dp    oversampling factor
    !   ioversampling   i8b   oversampling factor in int(i8b) form
    !   error           i8b   error number
    !   acTemp          char  temporary char array for output log messages
    !
            integer(i8b)                :: i, j, k, Ksource, Lsource, Msource, Lstart, Lend, Mstart, Mend, error
            real(dp), allocatable       :: xaxis(:), yaxis(:), xaxislrg(:), yaxislrg(:)
            complex(dpc), allocatable   :: sourceshapelrg(:, :), sourceshapesml(:, :)
            character(LEN=2048)         :: acTemp; 
            real(dp), parameter         :: oversampling = 8.0_dp
            integer(i8b)                :: ioversampling = oversampling
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   FFilterSpatial2D
    !   alloctest
    !   PrintToLog
    !
    ! =============================================================================
    
            write (acTemp, '("InitRectangularFlatAperture")'); call PrintToLog(acTemp, 3); 
            Ksource = size(fAperture, 1)
            Lsource = size(fAperture, 2)
            Msource = size(fAperture, 3)
            fAperture = 0.0_dp
    
            allocate (sourceshapelrg(1:ioversampling*Lsource, 1:ioversampling*Msource), &
                      sourceshapesml(1:Lsource, 1:Msource), &
                      xaxis(1:Lsource), yaxis(1:Msource), &
                      xaxislrg(1:ioversampling*Lsource), yaxislrg(1:ioversampling*Msource), STAT=error)
            call alloctest(error, 'Sourceshape and help variables in InitRectangularFlatAperture')
    
            !Axes normal size and large size in x and y
            xaxis = real((/(i, i=-Lsource/2, Lsource/2 - 1)/), dp)*dDx
            yaxis = real((/(i, i=-Msource/2, Msource/2 - 1)/), dp)*dDx
            xaxislrg = real((/(i, i=-ioversampling/2*Lsource, ioversampling/2*Lsource - 1)/), dp)*dDx/oversampling
            yaxislrg = real((/(i, i=-ioversampling/2*Msource, ioversampling/2*Msource - 1)/), dp)*dDx/oversampling
    
            ! Determine start and end indices of the rectangular source in x and y
            i = floor(cSourceParams%rectwidth/(2.0_dp*dDx/oversampling))
            Lstart = ioversampling/2*Lsource + 1 - i
            Lend = ioversampling/2*Lsource + 1 + i
            j = floor(cSourceParams%rectheight/(2.0_dp*dDx/oversampling))
            Mstart = ioversampling/2*Msource + 1 - j
            Mend = ioversampling/2*Msource + 1 + j
    
            ! Fill source array:
            ! 0.0 if outside source aperture
            ! 1.0 if on source aperture
            ! 0.5 if on the edge of the aperture
    
            sourceshapelrg = 0.0_dp
            sourceshapelrg(Lstart:Lend, Mstart:Mend) = 1.0_dp
    
            if ((xaxislrg(Lstart) + cSourceParams%rectwidth) < 1e-6) then
                sourceshapelrg(Lstart, Mstart:Mend) = 0.5_dp
                sourceshapelrg(Lend, Mstart:Mend) = 0.5_dp
            end if
    
            if ((yaxislrg(Mstart) + cSourceParams%rectheight) < 1e-6) then
                sourceshapelrg(Lstart:Lend, Mstart) = 0.5_dp
                sourceshapelrg(Lstart:Lend, Mend) = 0.5_dp
            end if
    
            !Filter the source spatially
            call FFilterSpatial2D(sourceshapelrg, sourceshapesml, dDx/oversampling, ioversampling, dKcutoff)
    
            do k = 1, Ksource
                fAperture(k, :, :) = sourceshapesml
            end do
    
            deallocate (sourceshapelrg, sourceshapesml, xaxis, yaxis, xaxislrg, yaxislrg)
    
        END SUBROUTINE InitRectangularFlatAperture
    
        SUBROUTINE InitTriangularFlatAperture(fAperture, dDx, dKcutoff)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitTriangularFlatAperture generates the aperture for the
    !   parametric description of a triangular source. The length of all edges
    !   is triangedge. The triangle is positioned symmetrically around the x-axis,
    !   with its top corner on the positive x-axis and its center on triangxcenter.
    !   The aperture is spatially filtered in X and Y.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   fAperture  o  dpc  3D array to be filled with the source aperture in the
    !                      (x,omega) domain.
    !   dDx        i  dp   Stepsize in x
    !   dKcutoff   i  dp   Cutoff spatial angular frequency
    !
            complex(dpc), intent(out) ::                    fAperture(:, :, :)
            real(dp), intent(in) ::                         dDx
            real(dp), intent(in) ::                         dKcutoff
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   i               i8b   temporary variable / loop counter
    !   j               i8b   temporary variable / loop counter
    !   k               i8b   temporary variable / loop counter
    !   ispan           i8b   span of the apertuer in x
    !   jspan           i8b   span of the apertuer in y
    !   icenter         i8b   center of the apertuer in x
    !   Ksource         i8b   size of the aperture in omega
    !   Lsource         i8b   size of the aperture in x
    !   Msource         i8b   size of the aperture in y
    !   xaxis           dp    x axis
    !   yaxis           dp    y axis
    !   xaxislrg        dp    x axis in the oversampling mode
    !   yaxislrg        dp    y axis in the oversampling mode
    !   sourceshapelrg  dpc   3D array of the source aperture in oversampling mode
    !   sourceshapesml  dpc   3D array containing the source aperture
    !   oversampling    dp    oversampling factor
    !   ioversampling   i8b   oversampling factor in int(i8b) form
    !   error           i8b   error number
    !   acTemp          char  temporary char array for output log messages
    !
            integer(i8b)                :: i, j, k, ispan, icenter, jspan, Ksource, Lsource, Msource, error
            real(dp), allocatable       :: xaxis(:), yaxis(:), xaxislrg(:), yaxislrg(:)
            complex(dpc), allocatable   :: sourceshapelrg(:, :), sourceshapesml(:, :)
            character(LEN=2048)         :: acTemp; 
            real(dp), parameter         :: oversampling = 8.0_dp
            integer(i8b)                :: ioversampling = oversampling
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   FFilterSpatial2D
    !   alloctest
    !   PrintToLog
    !
    ! =============================================================================
    
            write (acTemp, '("InitTriangularFlatAperture")'); call PrintToLog(acTemp, 3); 
            Ksource = size(fAperture, 1)
            Lsource = size(fAperture, 2)
            Msource = size(fAperture, 3)
            fAperture = 0.0_dp
    
            allocate (sourceshapelrg(1:ioversampling*Lsource, 1:ioversampling*Msource), &
                      sourceshapesml(1:Lsource, 1:Msource), &
                      xaxis(1:Lsource), yaxis(1:Msource), &
                      xaxislrg(1:ioversampling*Lsource), yaxislrg(1:ioversampling*Msource), STAT=error)
            call alloctest(error, 'Sourceshape and help variables in InitTriangularFlatAperture')
    
            !Axes normal size and large size in x and y
            xaxis = real((/(i, i=-Lsource/2, Lsource/2 - 1)/), dp)*dDx
            yaxis = real((/(i, i=-Msource/2, Msource/2 - 1)/), dp)*dDx
            xaxislrg = real((/(i, i=-ioversampling/2*Lsource, ioversampling/2*Lsource - 1)/), dp)*dDx/oversampling
            yaxislrg = real((/(i, i=-ioversampling/2*Msource, ioversampling/2*Msource - 1)/), dp)*dDx/oversampling
    
            ! Fill source array:
            ! 0.0 if outside source aperture
            ! 1.0 if on source aperture
            ! 0.5 if on the edge of the aperture
            sourceshapelrg = 0.0_dp
            ispan = floor(sqrt(3.0_dp)/4.0_dp*cSourceParams%triangedge/(dDx/oversampling))
            icenter = ioversampling*Lsource/2 + 1 + nint(cSourceParams%triangxcenter/(dDx/oversampling))
            do i = icenter - ispan, icenter + ispan
                jspan = floor((-(xaxislrg(i) - cSourceParams%triangxcenter)/sqrt(3.0_dp) + cSourceParams%triangedge/4.0_dp)/(dDx/oversampling))
                if ((xaxislrg(i) - cSourceParams%triangxcenter + sqrt(3.0_dp)/4.0_dp*cSourceParams%triangedge/(dDx/oversampling)) < 1e-6) then
                    do j = ioversampling*Msource/2 + 1 - jspan, ioversampling*Msource/2 + 1 + jspan
                        sourceshapelrg(i, j) = 0.5_dp
                    end do
                else
                    do j = ioversampling*Msource/2 + 1 - jspan, ioversampling*Msource/2 + 1 + jspan
                        sourceshapelrg(i, j) = 1.0_dp
                    end do
                end if
                ! If last points fall exactly on the border, fill them with 0.5_dp
                if (((xaxislrg(i) - cSourceParams%triangxcenter)/sqrt(3.0_dp) - cSourceParams%triangedge/4.0_dp) > -1e-6) then
                    sourceshapelrg(i, ioversampling*Msource/2 + 1 + jspan) = 0.5_dp
                    sourceshapelrg(i, ioversampling*Msource/2 + 1 - jspan) = 0.5_dp
                end if
            end do
    
    !call ExportArray(trim(sOutputDir) // trim(sOutputRoot) // "srclrg", "srclrg",dreal(sourceshapelrg), cSpace%cGrid%iProcID)
    
            !Filter the source spatially
            call FFilterSpatial2D(sourceshapelrg, sourceshapesml, dDx/oversampling, ioversampling, dKcutoff)
    
            do k = 1, Ksource
                fAperture(k, :, :) = sourceshapesml
            end do
    
            deallocate (sourceshapelrg, sourceshapesml, xaxis, yaxis, xaxislrg, yaxislrg)
    
        END SUBROUTINE InitTriangularFlatAperture
    
        SUBROUTINE InitPointSourceAperture(fAperture, dDx)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitPointSourceAperture generates the aperture for the
    !   parametric description of a point source. It is only the point at
    !   (x,y,z) = (0,0,0). Spatial filtering in x/y yields a sinc(x/y), which
    !   results in 1/dDx at 0 and zero elsewhere. Normalization with freq0/c0 of
    !   the point source is also required.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   fAperture  o  dpc  3D array to be filled with the source aperture in the
    !                      (x,omega) domain.
    !   dDx        i  dp   Stepsize in x
    !
            complex(dpc), intent(out) ::                    fAperture(:, :, :)
            real(dp), intent(in) ::                         dDx
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   acTemp          char  temporary char array for output log messages
    !
            character(LEN=2048) ::                          acTemp; 
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   PrintToLog
    !
    ! =============================================================================
    
            write (acTemp, '("InitPointSourceAperture")'); call PrintToLog(acTemp, 3); 
            !source shape is rather simple for point source
            ! account for factor (f0/c0) * 1/dx^2
            ! the first term is due to normalization of the equations
            ! and of the delta functions in x, y and z,
            ! the second term is the factor from 1/dx**2 sinc(\pi x / dx)*sinc(\pi y / dx) at 0,
            ! which is the result of the spatial filter in x and y
            ! By the way: the factor 1/dx from the filtering of the delta in z
            ! is cancelled by the convolution in z that is NOT performed in the
            ! evaluation of the convolution of the Green's function with the
            ! primary source.
            fAperture = 0.0_dp
            fAperture(:, 1, 1) = (cModelParams%freq0/cMediumParams%c0)*(1.0_dp/dDx)**2
    
        END SUBROUTINE InitPointSourceAperture
    
        SUBROUTINE InitPhasedArrayAperture(fAperture, dDx, dDw, dKcutoff)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitPhasedArrayAperture generates the aperture for the
    !   parametric description of a phased array. Focusing and steering in the
    !   scan plane (i.e. in x) is included with an elementwise phase delay, and
    !   focusing in the elevation direction (i.e. in y) is included over the
    !   elements with a phase delay depending on y.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   fAperture  o  dpc  3D array to be filled with the source aperture in the
    !                      (x,omega) domain.
    !   dDx        i  dp   Stepsize in x
    !   dDw        i  dp   Stepsize in omega
    !   dKcutoff   i  dp   Cutoff spatial angular frequency
    !
            complex(dpc), intent(out) ::        fAperture(:, :, :)
            real(dp), intent(in) ::                        dDx
            real(dp), intent(in) ::                        dDw
            real(dp), intent(in) ::                        dKcutoff
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   i               i8b   temporary variable / loop counter
    !   j               i8b   temporary variable / loop counter
    !   k               i8b   temporary variable / loop counter
    !   ispan           i8b   span of the apertuer in x
    !   jspan           i8b   span of the apertuer in y
    !   icenter         i8b   center of the apertuer in x
    !   Ksource         i8b   size of the aperture in omega
    !   Lsource         i8b   size of the aperture in x
    !   Msource         i8b   size of the aperture in y
    !   Mstart          i8b   start index in y
    !   Mend            i8b   end index in y
    !   phasedelay      dp    phase delay to account for elevation focusing in y
    !   temp            dp    temporary real variable
    !   si1             dp    sine integral used for the profile of an element in x
    !   si2             dp    sine integral used for the profile of an element in x
    !   sourcexprofile  dp    profile of an element in x
    !   waxis           dp    omega axis
    !   xaxis           dp    x axis
    !   yaxis           dp    y axis
    !   xaxislrg        dp    x axis in the oversampling mode
    !   yaxislrg        dp    y axis in the oversampling mode
    !   sourceapotemp   dp    temporary array to contain the apodization values
    !                         obtained from the file
    !   el_tdelay       dp    array with the delay times of the elements
    !   el_apo          dp    array with the apodization values of the elements
    !   el_x            dp    array with the positions in x of the elements
    !   el_term         dpc   array with the complex amplitude of one element
    !                         versus omega
    !   sourceyprofilelrg  dpc  2D array of the profile of an element in y
    !                           in oversampling mode
    !   sourceyprofilesml  dpc  2D array of the profile of an element in y
    !   oversampling    dp    oversampling factor
    !   ioversampling   i8b   oversampling factor in int(i8b) form
    !   error           i8b   error number
    !   acTemp          char  temporary char array for output log messages
    !
            integer(i8b)                    :: i, j, k, Ksource, Lsource, Msource, Mstart, Mend, error, YDimSource
            real(dp)                        :: phasedelay, temp, si1, si2, sourcexprofile, band
            real(dp), allocatable           :: waxis(:), xaxis(:), yaxis(:), xaxislrg(:), yaxislrg(:), &
                                                sourceapotemp(:), sourcetdtemp(:), el_tdelay(:), el_apox(:), el_apoy(:), el_x(:)
            complex(dpc), allocatable       :: sourceyprofilelrg(:, :), sourceyprofile(:, :), el_term(:)
            character(LEN=2048)             :: acTemp; 
            real(dp), parameter             :: oversampling = 8.0_dp
            integer(i8b)                    :: ioversampling = oversampling
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries and, if applicable, reading the apodization file
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   FFilterSpatial1D
    !   BINARY_read1Darraysize
    !   BINARY_read1Darray
    !   ASCII_read1Darraysize
    !   ASCII_read1Darray
    !   alloctest
    !   PrintToLog
    !
    ! *****************************************************************************
    !
    !   PSEUDOCODE
    !
    !   Initialize aperture, local variables, axes
    !   Determine the profile of a single element in y, in oversampling mode,
    !     including the phase delay that results in the elevation focusing
    !   Filter the oversampled y-profile spatially in y and reduce the sampling
    !   If applicable, read the apodization values of the elements from file
    !   Determine the x-positions and time delays of the elements
    !   For each element do
    !       Determine the complex amplitude = apodization * exp(-i * omega * delay)
    !       For each position in x do
    !           Determine the value of an analytically filtered block on this x
    !           For each frequency in omega do
    !               Add the y-line of the current element at this frequency and
    !                 x-position to the Aperture
    !   De-initialize local variables
    !
    ! =============================================================================
    
            write (acTemp, '("InitPhasedArrayAperture")'); call PrintToLog(acTemp, 3); 
            Ksource = size(fAperture, 1)
            Lsource = size(fAperture, 2)
            Msource = size(fAperture, 3)
            fAperture = 0.0_dp
    
            !Allocate help variables
            allocate (sourceyprofilelrg(1:Ksource, 1:ioversampling*Msource), &
                      sourceyprofile(1:Ksource, 1:Msource), &
                      el_tdelay(1:cSourceParams%numel), &
                      el_apox(1:cSourceParams%numel), &
                      el_apoy(1:Msource), &
                      el_x(1:cSourceParams%numel), el_term(1:Ksource), &
                      waxis(1:Ksource), xaxis(1:Lsource), yaxis(1:Msource), &
                      xaxislrg(1:ioversampling*Lsource), yaxislrg(1:ioversampling*Msource), STAT=error)
            call alloctest(error, 'Sourceshape and help variables in InitPhasedArrayAperture')
    
            !Axes in w and normal size and large size in x and y
            waxis = real((/(i, i=0, Ksource - 1)/), dp)*dDw
            xaxis = real((/(i, i=-Lsource/2, Lsource/2 - 1)/), dp)*dDx
            yaxis = real((/(i, i=-Msource/2, Msource/2 - 1)/), dp)*dDx
            xaxislrg = real((/(i, i=-ioversampling/2*Lsource, ioversampling/2*Lsource - 1)/), dp)*dDx/oversampling
            yaxislrg = real((/(i, i=-ioversampling/2*Msource, ioversampling/2*Msource - 1)/), dp)*dDx/oversampling
    
            !Determine y dimension depending on w - profile and delay
            Mstart = ioversampling/2*Msource + 1 - floor(cSourceParams%elheight/(2.0_dp*dDx/oversampling))
            Mend = ioversampling/2*Msource + 1 + floor(cSourceParams%elheight/(2.0_dp*dDx/oversampling))
    
            sourceyprofilelrg = 0.0_dp
            do j = Mstart, Mend
                if (abs(cSourceParams%elevationfocusz) > 1e-5) then
                    !maximum phasedelay at y=0, minimum at |y|=elheight/2
                    phasedelay = sqrt(yaxislrg(Mstart)**2 + cSourceParams%elevationfocusz**2) &
                                 - sqrt(yaxislrg(j)**2 + cSourceParams%elevationfocusz**2)
                else
                    phasedelay = 0.0
                end if
                sourceyprofilelrg(:, j) = exp(-im*waxis*phasedelay)
            end do
            if ((yaxislrg(Mstart) + cSourceParams%elheight) < 1e-6) then
                sourceyprofilelrg(:, Mstart) = 0.5_dp*sourceyprofilelrg(:, Mstart)
                sourceyprofilelrg(:, Mend) = 0.5_dp*sourceyprofilelrg(:, Mend)
            end if
    
            !filter y dimension by transform, filter, inverse transform and downsample
            call FFilterSpatial1D(sourceyprofilelrg, sourceyprofile, dDx/oversampling, ioversampling, dKcutoff)
    
            !x profile: determine elementwise, use analytically filtered elements
            !determine positions and time delays of elements
            el_x = -cSourceParams%arraywidth/2.0_dp + cSourceParams%elwidth/2.0_dp + &
                   (/(i, i=0, cSourceParams%numel - 1)/)*(cSourceParams%elwidth + cSourceParams%Kerf)
    
            !load time delay file
            if (trim(cSourceParams%td_filename) == 'none') then
                el_tdelay = -sqrt((cSourceParams%focusx - el_x)**2 + cSourceParams%focusz**2)*SIGN(1.0D0, cSourceParams%focusz)
            else
                i = index(cSourceParams%td_filename, '.')
                allocate (sourcetdtemp(1:cSourceParams%numel))
                if (cSourceParams%td_filename(i + 1:i + 3) == 'bin') then
                    if (BINARY_read1Darraysize(trim(sInputDir)//trim(cSourceParams%td_filename)) /= cSourceParams%numel) then
                        call PrintToLog("Error: number of items in apodization file is not equal to number of array elements", 1)
                        stop
                    end if
                    call BINARY_read1Darray(trim(sInputDir)//trim(cSourceParams%td_filename), sourcetdtemp)
                else
                    if (ASCII_read1Darraysize(trim(sInputDir)//trim(cSourceParams%td_filename)) /= cSourceParams%numel) then
                        call PrintToLog("Error: number of items in apodization file is not equal to number of array elements", 1)
                        stop
                    end if
                    call ASCII_read1Darray(trim(sInputDir)//trim(cSourceParams%td_filename), sourcetdtemp)
                end if
                el_tdelay = sourcetdtemp*cModelParams.freq0*1E-6; ! Normalize microsec
                deallocate (sourcetdtemp)
            end if
            el_tdelay = el_tdelay - minval(el_tdelay)
    
            !load apodization file
            if (trim(cSourceParams%phaseapodxfilename) == 'none') then
                el_apox = 1.0_dp
            else
                i = index(cSourceParams%phaseapodxfilename, '.')
                allocate (sourceapotemp(1:cSourceParams%numel))
                if (cSourceParams%phaseapodxfilename(i + 1:i + 3) == 'bin') then
                    if (BINARY_read1Darraysize(trim(sInputDir)//trim(cSourceParams%phaseapodxfilename)) /= cSourceParams%numel) then
                        call PrintToLog("Error: number of items in apodization file is not equal to number of array elements", 1)
                        stop
                    end if
                    call BINARY_read1Darray(trim(sInputDir)//trim(cSourceParams%phaseapodxfilename), sourceapotemp)
                else
                    if (ASCII_read1Darraysize(trim(sInputDir)//trim(cSourceParams%phaseapodxfilename)) /= cSourceParams%numel) then
                        call PrintToLog("Error: number of items in apodization file is not equal to number of array elements", 1)
                        stop
                    end if
                    call ASCII_read1Darray(trim(sInputDir)//trim(cSourceParams%phaseapodxfilename), sourceapotemp)
                end if
                el_apox = sourceapotemp
                deallocate (sourceapotemp)
            end if
    
            if (trim(cSourceParams%phaseapodyfilename) == 'none') then
                el_apoy = 1.0_dp
            else
                el_apoy = 0.0_dp
                YDimSource = ceiling(cSourceParams%elheight/(2.0_dp*dDx))
                band = 0.4_dp*(2*YDimSource + 1)*dDx/2.0_dp
                el_apoy(Msource/2 + 1 - YDimSource:Msource/2 + 1 + YDimSource) = 1.0_dp*dTaperingWindow(2*YDimSource + 1, dDx, band, band)
            end if
    
            !fill fAperture elementwise
            do i = 1, cSourceParams%numel
    
                !element excitation with apodization and focusing delay
                el_term = el_apox(i)*exp(-im*waxis*el_tdelay(i))
    
                !obtain (analytically) filtered x profile
                do j = 1, Lsource
                    call cisi(pi/dDx*(xaxis(j) - el_x(i) + cSourceParams%elwidth/2.0_dp), temp, si1)
                    call cisi(pi/dDx*(xaxis(j) - el_x(i) - cSourceParams%elwidth/2.0_dp), temp, si2)
                    sourcexprofile = (si1 - si2)/pi
    
                    do k = 1, Ksource
                        !add xy-line of the element to fAperture
                        fAperture(k, j, :) = fAperture(k, j, :) + &
                                             sourcexprofile*el_term(k)*sourceyprofile(k, :)*el_apoy
                    end do
                end do
    
            end do
    
            !Deallocate temp variables
            deallocate (sourceyprofilelrg, sourceyprofile, el_tdelay, el_apox, el_apoy, el_x, el_term, &
                        waxis, xaxis, yaxis, xaxislrg, yaxislrg)
    
        END SUBROUTINE InitPhasedArrayAperture
    
        SUBROUTINE InitMatrixArrayAperture(fAperture, dDx, dDw, dKcutoff)
    
    ! =============================================================================
    !
    !   Programmer: Koos Huijssen
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The subroutine InitPhasedArrayAperture generates the aperture for the
    !   parametric description of a phased array. Focusing and steering in the
    !   scan plane (i.e. in x) is included with an elementwise phase delay, and
    !   focusing in the elevation direction (i.e. in y) is included over the
    !   elements with a phase delay depending on y.
    !
    ! *****************************************************************************
    !
    !   INPUT/OUTPUT PARAMETERS
    !
    !   fAperture  o  dpc  3D array to be filled with the source aperture in the
    !                      (x,omega) domain.
    !   dDx        i  dp   Stepsize in x
    !   dDw        i  dp   Stepsize in omega
    !   dKcutoff   i  dp   Cutoff spatial angular frequency
    !
            complex(dpc), intent(out) ::                    fAperture(:, :, :)
            real(dp), intent(in) ::                         dDx
            real(dp), intent(in) ::                         dDw
            real(dp), intent(in) ::                         dKcutoff
    
    ! *****************************************************************************
    !
    !   LOCAL PARAMETERS
    !
    !   i               i8b   temporary variable / loop counter
    !   j               i8b   temporary variable / loop counter
    !   k               i8b   temporary variable / loop counter
    !   ispan           i8b   span of the apertuer in x
    !   jspan           i8b   span of the apertuer in y
    !   icenter         i8b   center of the apertuer in x
    !   Ksource         i8b   size of the aperture in omega
    !   Lsource         i8b   size of the aperture in x
    !   Msource         i8b   size of the aperture in y
    !   Mstart          i8b   start index in y
    !   Mend            i8b   end index in y
    !   phasedelay      dp    phase delay to account for elevation focusing in y
    !   temp            dp    temporary real variable
    !   si1             dp    sine integral used for the profile of an element in x
    !   si2             dp    sine integral used for the profile of an element in x
    !   sourcexprofile  dp    profile of an element in x
    !   waxis           dp    omega axis
    !   xaxis           dp    x axis
    !   yaxis           dp    y axis
    !   xaxislrg        dp    x axis in the oversampling mode
    !   yaxislrg        dp    y axis in the oversampling mode
    !   sourceapotemp   dp    temporary array to contain the apodization values
    !                         obtained from the file
    !   el_tdelay       dp    array with the delay times of the elements
    !   el_apo          dp    array with the apodization values of the elements
    !   el_x            dp    array with the positions in x of the elements
    !   el_term         dpc   array with the complex amplitude of one element
    !                         versus omega
    !   sourceyprofilelrg  dpc  2D array of the profile of an element in y
    !                           in oversampling mode
    !   sourceyprofilesml  dpc  2D array of the profile of an element in y
    !   oversampling    dp    oversampling factor
    !   ioversampling   i8b   oversampling factor in int(i8b) form
    !   error           i8b   error number
    !   acTemp          char  temporary char array for output log messages
    !
            integer(i8b)                :: i, j, k, Ksource, Lsource, Msource, Mstart, Mend, error
            real(dp)                    :: phasedelay, temp, si1, si2, sourcexprofile
            real(dp), allocatable       :: waxis(:), xaxis(:), yaxis(:), xaxislrg(:), yaxislrg(:), &
                                            sourceapotemp(:), el_tdelay(:), el_apo(:), el_x(:)
            complex(dpc), allocatable   :: sourceyprofilelrg(:, :), sourceyprofile(:, :), el_term(:)
            character(LEN=2048)         ::        acTemp; 
            real(dp), parameter         :: oversampling = 8.0_dp
            integer(i8b)                :: ioversampling = oversampling
    
    ! *****************************************************************************
    !
    !   I/O
    !
    !   Log file entries and, if applicable, reading the apodization file
    !
    ! *****************************************************************************
    !
    !   SUBROUTINES/FUNCTIONS CALLED
    !
    !   FFilterSpatial1D
    !   BINARY_read1Darraysize
    !   BINARY_read1Darray
    !   ASCII_read1Darraysize
    !   ASCII_read1Darray
    !   alloctest
    !   PrintToLog
    !
    ! *****************************************************************************
    !
    !   PSEUDOCODE
    !
    !   Initialize aperture, local variables, axes
    !   Determine the profile of a single element in y, in oversampling mode,
    !     including the phase delay that results in the elevation focusing
    !   Filter the oversampled y-profile spatially in y and reduce the sampling
    !   If applicable, read the apodization values of the elements from file
    !   Determine the x-positions and time delays of the elements
    !   For each element do
    !       Determine the complex amplitude = apodization * exp(-i * omega * delay)
    !       For each position in x do
    !           Determine the value of an analytically filtered block on this x
    !           For each frequency in omega do
    !               Add the y-line of the current element at this frequency and
    !                 x-position to the Aperture
    !   De-initialize local variables
    !
    ! =============================================================================
    
            write (acTemp, '("InitMatrixArrayAperture")'); call PrintToLog(acTemp, 3); 
    !KH beware: this subroutine still has to be implemented!!!
    !KH right now it is simply the code of InitPhaseArrayAperture
            write (acTemp, '("Not yet implemented")'); call PrintToLog(acTemp, 3); 
            stop
    
            Ksource = size(fAperture, 1)
            Lsource = size(fAperture, 2)
            Msource = size(fAperture, 3)
            fAperture = 0.0_dp
    
            !Allocate help variables
            allocate (sourceyprofilelrg(1:Ksource, 1:ioversampling*Msource), &
                      sourceyprofile(1:Ksource, 1:Msource), &
                      el_tdelay(1:cSourceParams%numel), &
                      el_apo(1:cSourceParams%numel), &
                      el_x(1:cSourceParams%numel), el_term(1:Ksource), &
                      waxis(1:Ksource), xaxis(1:Lsource), yaxis(1:Msource), &
                      xaxislrg(1:ioversampling*Lsource), yaxislrg(1:ioversampling*Msource), STAT=error)
            call alloctest(error, 'Sourceshape and help variables in InitPhasedArrayAperture')
    
            !Axes in w and normal size and large size in x and y
            waxis = real((/(i, i=0, Ksource - 1)/), dp)*dDw
            xaxis = real((/(i, i=-Lsource/2, Lsource/2 - 1)/), dp)*dDx
            yaxis = real((/(i, i=-Msource/2, Msource/2 - 1)/), dp)*dDx
            xaxislrg = real((/(i, i=-ioversampling/2*Lsource, ioversampling/2*Lsource - 1)/), dp)*dDx/oversampling
            yaxislrg = real((/(i, i=-ioversampling/2*Msource, ioversampling/2*Msource - 1)/), dp)*dDx/oversampling
    
            !Determine y dimension depending on w - profile and delay
            Mstart = ioversampling/2*Msource + 1 - floor(cSourceParams%elheight/(2.0_dp*dDx/oversampling))
            Mend = ioversampling/2*Msource + 1 + floor(cSourceParams%elheight/(2.0_dp*dDx/oversampling))
    
            sourceyprofilelrg = 0.0_dp
            do j = Mstart, Mend
                if (abs(cSourceParams%elevationfocusz) > 1e-5) then
                    !maximum phasedelay at y=0, minimum at |y|=elheight/2
                    phasedelay = sqrt(yaxislrg(Mstart)**2 + cSourceParams%elevationfocusz**2) &
                                 - sqrt(yaxislrg(j)**2 + cSourceParams%elevationfocusz**2)
                else
                    phasedelay = 0.0
                end if
                sourceyprofilelrg(:, j) = exp(-im*waxis*phasedelay)
            end do
            if ((yaxislrg(Mstart) + cSourceParams%elheight) < 1e-6) then
                sourceyprofilelrg(:, Mstart) = 0.5_dp*sourceyprofilelrg(:, Mstart)
                sourceyprofilelrg(:, Mend) = 0.5_dp*sourceyprofilelrg(:, Mend)
            end if
    
            !filter y dimension by transform, filter, inverse transform and downsample
            call FFilterSpatial1D(sourceyprofilelrg, sourceyprofile, dDx/oversampling, ioversampling, dKcutoff)
    
            !x profile: determine elementwise, use analytically filtered elements
    
            !load apodization file
            if (trim(cSourceParams%phaseapodxfilename) == 'none') then
                el_apo = 1.0_dp
            else
                i = index(cSourceParams%phaseapodxfilename, '.')
                allocate (sourceapotemp(1:cSourceParams%numel))
                if (cSourceParams%phaseapodxfilename(i + 1:i + 3) == 'bin') then
                    if (BINARY_read1Darraysize(trim(sInputDir)//trim(cSourceParams%phaseapodxfilename)) /= cSourceParams%numel) then
                        call PrintToLog("Error: number of items in apodization file is not equal to number of array elements", 1)
                        stop
                    end if
                    call BINARY_read1Darray(trim(sInputDir)//trim(cSourceParams%phaseapodxfilename), sourceapotemp)
                else
                    if (ASCII_read1Darraysize(trim(sInputDir)//trim(cSourceParams%phaseapodxfilename)) /= cSourceParams%numel) then
                        call PrintToLog("Error: number of items in apodization file is not equal to number of array elements", 1)
                        stop
                    end if
                    call ASCII_read1Darray(trim(sInputDir)//trim(cSourceParams%phaseapodxfilename), sourceapotemp)
                end if
                el_apo = sourceapotemp
            end if
    
            !determine positions and time delays of elements
            el_x = -cSourceParams%arraywidth/2.0_dp + cSourceParams%elwidth/2.0_dp + &
                   (/(i, i=0, cSourceParams%numel - 1)/)*(cSourceParams%elwidth + cSourceParams%Kerf)
            el_tdelay = -sqrt((cSourceParams%focusx - el_x)**2 + cSourceParams%focusz**2)
            el_tdelay = el_tdelay - minval(el_tdelay)
    
            !fill fAperture elementwise
            do i = 1, cSourceParams%numel
    
                !element excitation with apodization and focusing delay
                el_term = el_apo(i)*exp(-im*waxis*el_tdelay(i))
    
                !obtain (analytically) filtered x profile
                do j = 1, Lsource
                    call cisi(pi/dDx*(xaxis(j) - el_x(i) + cSourceParams%elwidth/2.0_dp), temp, si1)
                    call cisi(pi/dDx*(xaxis(j) - el_x(i) - cSourceParams%elwidth/2.0_dp), temp, si2)
                    sourcexprofile = (si1 - si2)/pi
    
                    do k = 1, Ksource
                        !add xy-line of the element to fAperture
                        fAperture(k, j, :) = fAperture(k, j, :) + &
                                             sourcexprofile*el_term(k)*sourceyprofile(k, :)
                    end do
                end do
    
            end do
    
            !Deallocate temp variables
            deallocate (sourceyprofilelrg, sourceyprofile, el_tdelay, el_apo, el_x, el_term, &
                        waxis, xaxis, yaxis, xaxislrg, yaxislrg)
    
        END SUBROUTINE InitMatrixArrayAperture
    
    #endif
    
    END MODULE ParnacDataSourceInit
    