% ***********************************************************************************************************
% Compare Analytic solution with simulation results
%
% ***************************************************************
% BubbleCluster_Compare(medium,domain,slice,file,plin,pnl,Bubble)
% ***************************************************************
% Compare the results generated from the matlab ODE solver with the ODEPACK Fortran90 solver
% for the Marmottant model , which describes how a micobubble behaves when excited by an acoustic wave
% Compare the results of the simulation with the analytic expression of the scattered pressure of one
% scatterer.
%
% Inputs:
% medium        struct      Contains all the medium properties, e.g Density, Speed of Sound etc
% domain        struct      Contains all the computational domain properties, e.g space and time step
% slice         struct      Contains all the information about the slice ,e.g Dimension, Position, Index etc
% file          struct      Contains all the loaded file properties ,e.g directory and root name etc
% plin          struct      Contains all the pressure values of the plin slice
% pnl           struct      Contains all the pressure values of the pnl slice
% Bubbles       struct      Contains all properties of the Scatterers' Cluster
%
% ******************
% AM, version 200723
% ******************
%
% ***********************************************************************************************************


function BubbleCluster_Compare_Individually(medium,domain,dslice,file,plin,pnl,Bubble)
disp(' ')
disp('Calculating the difference between analytic and simulation solution...')
if (Bubble.N <=10 && Bubble.N>0)
    
    Sim_OSFactor=1;
    %%
    lin_p = plin.data;%plin.data_40{10};
    nonlin_p= pnl.data;%nl.data_40{10,10};
    con_p = (nonlin_p-lin_p);
    
    i_start = 1;
    i_end = size(lin_p,1)-200;
    depth = 70; %depth =  55;
    x_idx = round(size(lin_p,2)/2)+1; %x_idx=47;
    figure('WindowState','maximized');
    tpar = domain.tpar(i_start:i_end)*1E6;
    hold on;  plot([1:length(tpar)*10]*(tpar(2)-tpar(1))/10,resampleSINC(nonlin_p(i_start:i_end,x_idx,depth)'*1E-3,10), '-','LineWidth',3)
    hold on;  plot([1:length(tpar)*10]*(tpar(2)-tpar(1))/10,resampleSINC(lin_p(i_start:i_end,x_idx,depth)'*1E-3,10),'-.','LineWidth',3)
    hold on;  plot([1:length(tpar)*10]*(tpar(2)-tpar(1))/10,resampleSINC(con_p(i_start:i_end,x_idx,depth)'*1E-3,10), '--','LineWidth',3)
    xlabel('\textit{t} [$\mu$s]')
    ylabel('\textit{p} [kPa]')
    set(gca,'FontSize',20)
    set(gcf,'Color','white')
    hl = legend('Total','Linear','Contrast');
    grid on;
    title(['INCS Results at (X,Y,Z) = (', num2str(domain.par{1}(x_idx)),', ',num2str(dslice.pos(dslice.num)),', ', num2str(domain.par{3}(depth)),') [mm]'])
    
    F_s = 1/(domain.dtpar);                          % [Hz] , 1/dt
    [f_s,Sp_t] = Freq_Calc(nonlin_p(i_start:i_end,x_idx,depth),F_s);            % Spectrum of simulation
    [f_p,Sp_p] = Freq_Calc(lin_p(i_start:i_end,x_idx,depth),F_s);            % Spectrum of simulation
    [f_s,Sp_s] = Freq_Calc(con_p(i_start:i_end,x_idx,depth),F_s);            % Spectrum of simulation
    figure('WindowState','maximized');
    hold on;  plot(f_s*1E-6,Sp_t,'-','LineWidth',3)
    hold on;  plot(f_p*1E-6,Sp_p ,'-.','LineWidth',3)
    hold on;  plot(f_s*1E-6,Sp_s,'--','LineWidth',3)
    ylim(max([Sp_p ; Sp_t]) - [ 100 -5])
    xlim([0 domain.Fnyq*medium.freq0*1E-6])
    xlabel('\textit{f} [MHz]')
    ylabel('Amplitude [dB]')
    set(gca,'FontSize',20)
    set(gcf,'Color','white')
    legend('Total','Linear','Scattered')
    grid on;
%     ylim([max([Sp_p; Sp_s])-70 max([Sp_p; Sp_s])])
    %% ============== Comparison of time signature based on ODE Solvers ====================================
    if (strcmp(file.scatterer,'active'))
        %%
        % Load pressure values from file
        %             filename_pres = filename_pres(1:end-1);
        filename_pres = split(cellstr(ls([file.dirname '/Bubbles/' 'v_dd_norm_001*.txt'])));
        for ifile = 1:1
            fileID_pres = fopen([file.dirname '/Bubbles/' filename_pres{ifile} ],'r');
            size_pres = [1 Inf];
            formatSpec = '%f';
            
            V_dd_sim_loaded = fscanf(fileID_pres,formatSpec,size_pres);
            
            n_pad = (length(V_dd_sim_loaded)/domain.tdimpar-2)/2;
            n_pad = 2;
            tdimpar_file = domain.tdimpar;%size(Bubble.Contrast,2);
            iBubble_ofinterest =  1;
%             if (length(V_dd_sim_loaded)>(1+2*domain.tdimpar+2*n_pad*domain.tdimpar)); iBubble_ofinterest = 1;end
            V_dd_sim = V_dd_sim_loaded((iBubble_ofinterest-1)*(2*tdimpar_file+2*n_pad*tdimpar_file) + 1 :iBubble_ofinterest*(2*tdimpar_file+2*n_pad*tdimpar_file));
            iBubble = V_dd_sim_loaded((iBubble_ofinterest-1)*(2*tdimpar_file+2*n_pad*tdimpar_file) + 1)
            
            V_dd_Time = V_dd_sim(1:tdimpar_file);
            V_dd_Pres = V_dd_sim(tdimpar_file + 1 : 2*tdimpar_file);
            V_dd_TimePad = V_dd_sim(2*tdimpar_file + 1:(n_pad+2)*tdimpar_file);
            V_dd_PresPad = V_dd_sim((n_pad+2)*tdimpar_file + 1 : (n_pad+2)*tdimpar_file+n_pad*tdimpar_file);
            
            figure('WindowState','maximized');
            hold on; plot(V_dd_Time*1E6,V_dd_Pres,V_dd_TimePad*1E6,V_dd_PresPad)
            legend('Initial Pulse', 'Resampled Pulse')
            set(gca,'FontSize',20)
            set(gcf,'Color','white')
            xlabel('Time [usec]')
            ylabel('Contrast Source term [Pa/m^2]')
            
            F_s = 1/(domain.dtpar/n_pad);                          % [Hz] , 1/dt
            [f_p,Sp_p] = Freq_Calc(V_dd_Pres,1/(domain.dtpar));            % Spectrum of simulation
            [f_ppad,Sp_ppad] = Freq_Calc(V_dd_PresPad,F_s);                % Spectrum of analytical pulse
            figure('WindowState','maximized');
            hold on;  plot(f_p*1E-6,Sp_p-max(Sp_p),f_ppad*1E-6,Sp_ppad-max(Sp_ppad))
            xlim([0 1/2/domain.dtpar*1E-6])
            legend('Initial Pulse', 'Resampled Pulse')
            set(gca,'FontSize',20)
            set(gcf,'Color','white')
            xlabel('Frequency [MHz]')
            ylabel('Normalized amplitude [dB]')
            
            % Create Simulation Contrast, correct if n_padded in INCS code
            BubbleTimeSim = resampleSINC(domain.tpar,n_pad);
            Sim_OSFactor = n_pad*length(Bubble.Contrast(1,:))/length(BubbleTimeSim);
            if Sim_OSFactor>1 ; BubbleCon_Sim = downsample(Bubble.Contrast(1,:),Sim_OSFactor/n_pad); else ; BubbleCon_Sim = resampleSINC(Bubble.Contrast(1,:),n_pad/Sim_OSFactor); end
        end
        fclose('all')
        %%
        filename_pres = split(cellstr(ls([file.dirname '/Bubbles/output_file_pressure_001*.txt'])));
        for ifile = 1:1
            fileID_pres = fopen([file.dirname '/Bubbles/' filename_pres{ifile} ],'r');
            size_pres = [1 Inf];
            formatSpec = '%f';
            
            BubbleVal_loaded = fscanf(fileID_pres,formatSpec,size_pres);
         
            n_pad = 2;
            tdimpar_file = domain.tdimpar;%size(Bubble.Contrast,2);
            iBubble_ofinterest = 1;
%             if (length(BubbleVal_loaded)>(1+2*tdimpar_file+2*n_pad*tdimpar_file)); iBubble_ofinterest = 20;end
            BubbleVal = BubbleVal_loaded((iBubble_ofinterest-1)*(1+2*tdimpar_file+2*n_pad*tdimpar_file) + 2 :iBubble_ofinterest*(1+2*tdimpar_file+2*n_pad*tdimpar_file));
            iBubble = BubbleVal_loaded((iBubble_ofinterest-1)*(1+2*tdimpar_file+2*n_pad*tdimpar_file) + 1)
            
            BubbleTime = BubbleVal(1:tdimpar_file);
            BubblePres = BubbleVal(tdimpar_file + 1 : 2*tdimpar_file);
            BubbleTimePad = BubbleVal(2*tdimpar_file + 1:(n_pad+2)*tdimpar_file);
            BubblePresPad = BubbleVal((n_pad+2)*tdimpar_file + 1 : (n_pad+2)*tdimpar_file + n_pad*tdimpar_file);
            figure('WindowState','maximized');
            hold on; plot(BubbleTime*1E6,BubblePres,BubbleTimePad*1E6,BubblePresPad)
            legend('Initial Pulse', 'Resampled Pulse')
            set(gca,'FontSize',20)
            set(gcf,'Color','white')
            xlabel('Time [usec]')
            ylabel('Pressure [Pa]')
            grid on;
            
            F_s = 1/(domain.dtpar/n_pad);                          % [Hz] , 1/dt
            [f_p,Sp_p] = Freq_Calc(BubblePres,1/domain.dtpar);            % Spectrum of simulation
            [f_ppad,Sp_ppad] = Freq_Calc(BubblePresPad,F_s);                % Spectrum of analytical pulse
            figure('WindowState','maximized');
            hold on;  plot(f_p*1E-6,Sp_p-max(Sp_p),f_ppad*1E-6,Sp_ppad-max(Sp_ppad))
            legend('Initial Pulse', 'Resampled Pulse')
            set(gca,'FontSize',20)
            set(gcf,'Color','white')
            xlabel('Frequency [MHz]')
            ylabel('Normalized amplitude [dB]')
            xlim([0 1/2/domain.dtpar*1E-6])
            grid on;
            
            % Create Simulation Contrast, correct if n_padded in INCS code
            BubbleTimeSim = resampleSINC(domain.tpar,n_pad);
            Sim_OSFactor = n_pad*length(Bubble.Contrast(1,:))/length(BubbleTimeSim);
            if Sim_OSFactor>1 ; BubbleCon_Sim = downsample(Bubble.Contrast(1,:),Sim_OSFactor/n_pad); else ; BubbleCon_Sim = resampleSINC(Bubble.Contrast(1,:),n_pad/Sim_OSFactor); end
        end
        fclose('all')
        %%
        close all
%         BubblePresPad = resampleSINC(plin.data(:,72,80)',n_pad);
        BubbleTimeAn = BubbleTimePad;
        [par, T_m, V_dd_norm] = Bubble_SimAnalytic(BubblePres,BubbleTime,medium.c0,medium.rho0,medium.freq0,domain);         % Solution of RP Solver with ode45 MATLAB solver
        BubbleCon_an = V_dd_norm*medium.rho0;                                                             % Contrast Source Term
        %========================== Save plot ===================================
        if (strcmp(file.saveplot,'yes'))
            saveas(gcf,[file.savedir,'/',file.dirname,'_bubble_motion_slice',int2str(dslice.num),'.jpg'])
        end
        %
        f5 = figure('WindowState','maximized');
        ax = gca; hold on;
        plot(BubbleTime*1E6,medium.rho0*V_dd_Pres,'k','LineWidth',1.5);
        plot(BubbleTime*1E6,BubbleCon_an,'b--','LineWidth',1.5);
        legend('Simulation', 'Matlab ODE')
        set(gca,'FontSize',20)
        set(gcf,'Color','white')
        xlabel('$\textit{t}$ [$\mu$s]')
        ylabel('Contrast Source term [Pa/m$^2$]')
        grid on;
        
        F_s = 1/(domain.dtpar/n_pad);                          % [Hz] , 1/dt
        [f_p,Sp_p] = Freq_Calc(BubbleCon_an,1/(domain.dtpar));            % Spectrum of simulation
        [~,Sp_ppad] = Freq_Calc(medium.rho0*V_dd_Pres,1/(domain.dtpar));                % Spectrum of analytical pulse
        f_p = f_p*1E-6/n_pad;
        figure('WindowState','maximized');
        hold on;
        plot(f_p(1:tdimpar_file),Sp_p(1:tdimpar_file),'k','LineWidth',1.5);
        plot(f_p(1:tdimpar_file),Sp_ppad(1:tdimpar_file),'b--','LineWidth',1.5)
        set(gca,'FontSize',20)
        set(gcf,'Color','white')
        xlabel('\textit{f} [MHz]')
        ylabel('$\hat{p}$ [dB]')
        legend('Simulation', 'Matlab ODE')
        grid on;
        fclose('all')
    end
    
    %% =============================================== Compare simulation with analutical Results =======================================================
    % This is working for 1 or 2 Bubbles and up to 1st order multiple scattering
    % it can easily expanded to more but these are enough for the validation of the code
    %         close all
    
    OSFactor = 	1;
    R_rigid = repmat(2E-6,Bubble.N,1); %R_rigid(1) = 0;
    time_signature  =   (4/3*pi*R_rigid.^3)/(medium.c0^2);
    medium.rho1     =   10;
    medium.c1       =   100;
    time_signature  =   -(4/3*pi*R_rigid.^3)*medium.rho0.*(1./(medium.rho1*medium.c1.^2)- 1/(medium.rho0*medium.c0^2));
    time_signature = V_dd_norm*medium.rho0;
    %======== Driving Frequency ======================
    freq        = medium.freq0;
    w_driving   = 2*pi*freq;
    gauss_dl    = 6/freq;
    gauss_win   = 3/freq;
      
    domain.tpar_OS = interp(domain.tpar,OSFactor);
    domain.dtpar_OS = domain.tpar_OS(2) - domain.tpar_OS(1);
    domain.tdimpar_OS = length(domain.tpar_OS);
    
    Inc_Press   = repmat(2E5*exp( - ((domain.tpar_OS-gauss_dl)/(gauss_win/2)).^2).* sin(w_driving*(domain.tpar_OS-gauss_dl)),Bubble.N,1);
    Fs          = 1/domain.dtpar_OS;  dF = Fs/domain.tdimpar_OS;
    if mod(domain.tdimpar_OS,2) == 0 % Nt is even
        f = -Fs/2 : dF : Fs/2-dF;
    else            % Nt is odd
        f = [sort(-1*(dF:dF:Fs/2)) (0:dF:Fs/2)];
    end
    omega       = 2*pi*f;
    omega       = [omega(omega>=0) omega(omega<0)];
    omega       = repmat(omega,Bubble.N,1);
    
    % This is only for the case of planewave. This is a time shift for the travelling of a planewave.
    Inc_Press = real(ifft(fft(Inc_Press,[],2).*exp(-1i*omega/medium.c0.*Bubble.LocGlob(:,3)*1E-3),[],2));
    % Make denser the matrix to increase accuracy
%     pdiff  = pnl.contrastdata;  % If you do not save contrastdata,  pdiff = plin.data - pnl.data
%     
    %======= Initialize Arrays ========================
    pnl.simN                = zeros(domain.tdimpar_OS*OSFactor,1);
    pnl.anN                 = zeros(domain.tdimpar_OS*OSFactor,1);
    pnl.semi_anN            = zeros(domain.tdimpar_OS*OSFactor,1);
    pnl.anN_map             = zeros(domain.tdimpar_OS*OSFactor,1);
    
    [~,SliceIndex(2)] = min(abs(dslice.pos(dslice.num)-domain.par{domain.dimval(1)}));
    [f_p,Sp_p] = Freq_Calc(Inc_Press(1,:),Fs);            % Spectrum of simulation
    % Iterate over k in x axis and m in z axis. This implementation is done for a y slice so the index in y equals always to the closest point of the slice position (SliceIndex(2))
    %% Generate the pressure that each microbubble understands at the iteration of interest
    P_scattered = zeros(Bubble.N,domain.tdimpar_OS);
    TotalPress = Inc_Press;
    LocGlob = Bubble.LocGlob;
    for j = 1: domain.beamiterations-1
        P_scattered = zeros(Bubble.N,domain.tdimpar_OS);
        for LocBubble = 1:ceil(Bubble.N/2)
            r_from_scatterer = sqrt((Bubble.LocGlob(LocBubble,1) - Bubble.LocGlob(:,1)).^2 + (Bubble.LocGlob(LocBubble,2) - Bubble.LocGlob(:,2)).^2 + (Bubble.LocGlob(LocBubble,3) - Bubble.LocGlob(:,3)).^2)*1e-3;
            Greens_function = exp(-1i*omega.*r_from_scatterer/medium.c0)./(4*pi*r_from_scatterer); % 1/(4πr) and the time shift due to wave travel
            Greens_function(isinf(Greens_function)) = 0;
            
            ddpddt      =   fft(TotalPress(LocBubble,:),[],2).*(1i*omega(LocBubble,:)).^2 ; % What the scatterer understands
            an      =   ddpddt.* time_signature(LocBubble) .* Greens_function;
            an      =   real(ifft(an,[],2));
            
            iBubble_end = Bubble.N - LocBubble+1;
            r_from_scatterer_end = sqrt((Bubble.LocGlob(iBubble_end,1) - Bubble.LocGlob(:,1)).^2 + (Bubble.LocGlob(iBubble_end,2) - Bubble.LocGlob(:,2)).^2 + (Bubble.LocGlob(iBubble_end,3) - Bubble.LocGlob(:,3)).^2)*1e-3;
            Greens_function_end = exp(-1i*omega.*r_from_scatterer_end/medium.c0)./(4*pi*r_from_scatterer_end); % 1/(4πr) and the time shift due to wave travel
            Greens_function_end(isinf(Greens_function_end)) = 0;
            
            ddpddt_end  =   fft(TotalPress(iBubble_end,:),[],2).*(1i*omega(iBubble_end,:)).^2 ; % What the scatterer understands
            an_end  =   ddpddt_end.* time_signature(iBubble_end) .* Greens_function_end;
            an_end  =   real(ifft(an_end,[],2));
            
            % This array contains for each row, the pressure that this bubble understands as a summation of all its neighbors
            P_scattered = P_scattered + an + an_end;
        end
        TotalPress      =   Inc_Press + P_scattered;
        
    end
    %% Do the same only for one bubble
    P_scattered = zeros(Bubble.N,domain.tdimpar_OS);
    TotalPress = Inc_Press(iBubble,:);
    Greens_function = 0;
    for j = 1: domain.beamiterations-1
        P_scattered = zeros(1,domain.tdimpar_OS);
        for LocBubble = 1:ceil(Bubble.N/2)
            r_from_scatterer = sqrt((Bubble.LocGlob(LocBubble,1) - Bubble.LocGlob(iBubble,1)).^2 + (Bubble.LocGlob(LocBubble,2) - Bubble.LocGlob(iBubble,2)).^2 + (Bubble.LocGlob(LocBubble,3) - Bubble.LocGlob(iBubble,3)).^2)*1e-3;
            Greens_function = exp(-1i*omega(iBubble,:).*r_from_scatterer/medium.c0)./(4*pi*r_from_scatterer); % 1/(4πr) and the time shift due to wave travel
            Greens_function(isinf(Greens_function)) = 0;
            
            ddpddt      =   fft(TotalPress,[],2).*(1i*omega(LocBubble,:)).^2 ; % What the scatterer understands
            an      =   ddpddt.* time_signature(LocBubble) .* Greens_function;
            an      =   real(ifft(an,[],2));
            
            iBubble_end = Bubble.N - LocBubble+1;
            r_from_scatterer_end = sqrt((Bubble.LocGlob(iBubble_end,1) - Bubble.LocGlob(iBubble,1)).^2 + (Bubble.LocGlob(iBubble_end,2) - Bubble.LocGlob(iBubble,2)).^2 + (Bubble.LocGlob(iBubble_end,3) - Bubble.LocGlob(iBubble,3)).^2)*1e-3;
            Greens_function_end = exp(-1i*omega(iBubble,:).*r_from_scatterer/medium.c0)./(4*pi*r_from_scatterer);
            Greens_function_end(isinf(Greens_function_end)) = 0;
            
            ddpddt_end  =   fft(TotalPress,[],2).*(1i*omega(iBubble_end,:)).^2 ; % What the scatterer understands
            an_end  =   ddpddt_end.* time_signature(iBubble_end) .* Greens_function_end;
            an_end  =   real(ifft(an_end,[],2));
            
            % This array contains for each row, the pressure that this bubble understands as a summation of all its neighbors
            P_scattered = P_scattered + an + an_end;
        end
        TotalPress      =   Inc_Press(iBubble,:) + P_scattered;
        
    end
    %%
    k_init      =   1;%79;
    k_end       =   domain.dimlen(1);
    m_init      =   1;%domain.dimlen(3);
    m_end       =   domain.dimlen(3);
    % % %
%     %
%     m_init = idx1;
%     k_init = idx3;
%     m_init      =   42;49;77;
%     m_end       =   m_init;
%     k_init      =   30;11;58;
%     k_end       =   k_init;
% %         %
    RRMS_ErrorG_semi_an = zeros(k_end,m_end);
    RRMS_ErrorG_an = zeros(k_end,m_end);
    for i = domain.beamiterations:domain.beamiterations
        TotalPress_lastiter      =   TotalPress;
        ddpddt          =   fft(TotalPress_lastiter,[],2).*(1i*omega).^2 ; % What the scatterer understands
         %======= Initialize Arrays ========================
        % Initialize for higher order multiple scattering . This is done because on previous j's scattered pressure is calculated
        pnl.simN_cluster        = zeros(domain.tdimpar*OSFactor,1);
        pnl.anN_cluster         = zeros(domain.tdimpar*OSFactor,domain.dimlen(1),domain.dimlen(3));
        pnl.semi_anN_cluster    = zeros(domain.tdimpar*OSFactor,domain.dimlen(1),domain.dimlen(3));
        for k = k_init: k_end
            for m = m_init:m_end
                % Iterate to include higher orders of multiple scattering
                % Find the position of interest which is the position of slice
                % Find the grid points with which the bubble has the smallest distance
                SliceIndex(1) = k ;  SliceIndex(3) = m;
                %================================================== GLOBAL VARIABLES =============================================================
                % This is used to show the propagation in the horizontal axis,This is because of the results of INCS
                % As in the movies, there should be a shift in
                r_from_scatterer = sqrt((Bubble.LocGlob(:,domain.dimval(2)) - domain.par{domain.dimval(2)}(SliceIndex(1))).^2+(Bubble.LocGlob(:,domain.dimval(1)) - dslice.pos(dslice.num)).^2 ...
                    +(Bubble.LocGlob(:,domain.dimval(3))-domain.par{domain.dimval(3)}(SliceIndex(3))).^2)*1e-3;
                z_shift = domain.par{domain.dimval(3)}(SliceIndex(3)) *1e-3;

                Greens_function = (exp(-1i*omega.*r_from_scatterer/medium.c0)./(4*pi*r_from_scatterer)); % 1/(4πr) and the time shift due to wave travel
                Greens_function(isinf(Greens_function)) = 0;
                
                %================================================== Compare for passive scattering =======================================
                % The shift due to the time that takes for the planewave to travel to the position of the scatterer
                % is not taken into account because the output of INCS do not include this
                % that's why in INCS the movie should be shifted with the z axis, otherwise the result is not logical.
                % but in case of test, r_from_transducer = sqrt(sum(Bubble.LocGlob(iBubble,3).^2))*1e-3;
                
                %Spectral derivative
             
%                 ddpddt_lastiter = ddpddt.* time_signature .* Greens_function;
                ddpddt_lastiter = fft(time_signature,[],2) .* Greens_function;
                pnl.an = ddpddt_lastiter;
                pnl.an = real(ifft(pnl.an,[],2));
                pnl.anN_cluster(:,k,m) = sum(pnl.an,1)';
                
                %================================================== SEMI-ANALYTIC RESULTS SHIFTED =============================================================
                % This phase shift is due to the shift that was implemented in the incident pressure field in the beginning
                % When a comoving time window is used then all the results need to be shifted in the correct position
%                 pnl.semi_an = [zeros(1,domain.tdimpar - length(Bubble.Contrast)) Bubble.Contrast] ; % Semi - analytic because checking only the Green's function
%                 pnl.semi_an = real(ifft(fft(pnl.semi_an,[],2).*Greens_function,[],2));
%                 pnl.semi_anN_cluster(:,k,m) = sum(pnl.semi_an,1)';%interp(pnl.semi_an,OSFactor)'; % Normalize to calculate error
                
                %================================================== SIMULATION RESULTS =============================================================
                pnl.sim = squeeze(pdiff(:,SliceIndex(1),SliceIndex(3)));
%                 if (domain.a_t);pnl.sim = real(ifft(fft(pnl.sim').* exp(-1i*omega(1,:)* (z_shift(1,1)/medium.c0 - domain.tstart*domain.dtpar) )))';end
                % The check for the time signature for microbubbles is done from the previous step of the comparison of the ode solvers
                pnl.simN_cluster(:,k,m) = pnl.sim;%interp(pnl.sim,OSFactor); % Normalize to calculate error
                
                %Safety measure , max 10 plots
%                 if ( (k_end-k_init+1)*(m_end-m_init+1)<10)
%                     Plot_Error(domain.tpar_OS,pnl,domain,SliceIndex,RRMS_ErrorG_semi_an,RRMS_ErrorG_an,k,m,Bubble.N,i)
%                     Plot_Spectral_Response(domain,pnl,OSFactor,k,m)
%                     %% ========================== Save plot ===================================
%                     if (strcmp(file.saveplot,'yes'))
%                         saveas(gca,[file.savedir,'/',file.dirname,'_comparison_scattered_pressure_slice',int2str(dslice.num),int2str([i;j;iBubble])','.jpg'])
%                         %saveas(f7,[file.savedir,'/',file.dirname,'_frequencyspectrum',int2str(dslice.num),int2str([i;j])','.jpg'])
%                     end
%                 end
            end
        end
    end
    
    %% ========================================= PLOT ERROR IN 2D SLICE ==================================================================
    %     if ( k_end == domain.dimlen(1) && m_end == domain.dimlen(3))
    RRMS_Err_Corr = RRMS_ErrorG_an;
%     RRMS_Err_Corr(abs(RRMS_Err_Corr)>0.16)=0;
%         RRMS_Err_Corr(abs(RRMS_Err_Corr)>0.2)=0;
    figure('WindowState','maximized');
    ax1 = axes;
    length_z = 1;
    length_x = 1;
    contourf(domain.par{3}(length_z:end-length_z-1),domain.par{1}(length_x:end-length_x-1),RRMS_Err_Corr(length_x:end-length_x-1,length_z:end-length_z-1)*100);
%     contourf(RRMS_Err_Corr*100)
%     RRMS_Err_Corr(abs(RRMS_Err_Corr)>=0.1)=0.1;
%     RRMS_Err_Corr(abs(RRMS_Err_Corr)>=0.01 & abs(RRMS_Err_Corr)<0.1)=0.01;
%     RRMS_Err_Corr(abs(RRMS_Err_Corr)>=0.001 & abs(RRMS_Err_Corr)<0.01)=0.001;
%     RRMS_Err_Corr(abs(RRMS_Err_Corr)<0.01)=0;
%     [C, h] = contourf(domain.par{3}(length_z:end-length_z-1),domain.par{1}(length_x:end-length_x-1),RRMS_Err_Corr(length_x:end-length_x-1,length_z:end-length_z-1)*100,[0.1 1 10],'Color','black'); 
%     clabel(C,h)
%     xlim([2 8]); ylim([-3 3])
    xlabel('Z [mm]')
    ylabel('X [mm]')
    c1 = colorbar(ax1);
    c1.Label.String = 'RRRMS [%]';
    set(gca,'FontSize',20)
    hold on;
        if (Bubble.N >6)
            rectangle('Position',[Bubble.min_dim(domain.dimval(3)) Bubble.min_dim(domain.dimval(2)) Bubble.max_dim(domain.dimval(3))-Bubble.min_dim(domain.dimval(3)) Bubble.max_dim(domain.dimval(2))-Bubble.min_dim(domain.dimval(2))],'LineStyle','--','EdgeColor','white','LineWidth',5)
        else
            plot(ax1,Bubble.LocGlob(:,3),Bubble.LocGlob(:,1), 'x','Color','white','MarkerSize',20,'LineWidth',4)
        end
    BubbleCluster_Colormaps(file)
    shading interp;
    
    title(['Analytic vs Simulation of MB scattered pressure at y = ',num2str(dslice.pos(dslice.num)),])
    mean(mean(RRMS_Err_Corr))
    std(std(RRMS_Err_Corr))
    set(gca,'FontSize',20)
    set(gcf,'Color','white');
    %%
    k = 39; m =35;
    pnl.simN_cluster(:,k,m) = squeeze(pdiff(:,k,m));
    [RRMS_ErrorG_an(k,m),Rel_ErrorG_an(k,m)] = Error(pnl.anN_cluster(:,k,m),pnl.simN_cluster(:,k,m));
%                 [RRMS_ErrorG_semi_an(k,m),Rel_ErrorG_semi_an(k,m)] = Error(pnl.semi_anN_cluster(:,k,m),pnl.simN_cluster(:,k,m));
    Plot_Error(domain.tpar_OS,pnl,domain,SliceIndex,RRMS_ErrorG_semi_an,RRMS_ErrorG_an,k,m,Bubble.N,i)
    Plot_Spectral_Response(domain,pnl,OSFactor,k,m)
    %% ========================================= PLOT Scattered pressure ==================================================================
    
    for iBubble = 1:Bubble.N
        
        figure;
        ax = gca;
        hold on;
        plot(domain.tpar_OS,P_scattered(iBubble,:))
        ax=gca;
        xlabel('time [μsec]')
        set(gca,'XTickLabel',ax.XTick*1e+6) ; ax.XLabel.String = 'Time [μsec]';
        ylabel('Pressure [Pa]')
        title(['Scattered pressure of Bubble No. ',num2str(Bubbles(iBubble)),' to Bubble No. ' num2str(iBubble), '.'])
        hold off;
    end
end
end

%% Calculate RRMS and Relative Error between analytic and simulation
function [RRMS_ErrorG_an,Rel_ErrorG_an] = Error(p_an,p_sim)

RRMS_ErrorG_an = sqrt( sum((p_sim - p_an).^2,1)./sum((p_an).^2 ));

[~,i_maxdiff] = max(abs(p_sim-p_an));
Rel_ErrorG_an = sqrt( sum((p_sim - p_an).^2,1)./sum((p_an-max(p_an)).^2 )); % Max as an estimator


end

%% ================================================ Plot Error================================================

function Plot_Error(tpar_N,pnl,domain,SliceIndex,RRMS_ErrorG_semi_an,RRMS_ErrorG_an,k,m,iBubble,i)
f6 = figure('WindowState','maximized');
h_f1 = subplot(2,1,1);
ax = gca; hold on;
p1 = plot(tpar_N*1E6 ,pnl.simN_cluster(:,k,m),'--o','MarkerIndices',1:50:length(tpar_N),'MarkerSize',4,'color','k');
p3 = plot(tpar_N*1E6 ,pnl.semi_anN_cluster(:,k,m),':x','MarkerIndices',1:50:length(tpar_N),'MarkerSize',4,'color',[17 17 17]./255);
p2 = plot(tpar_N*1E6 ,pnl.anN_cluster(:,k,m),'-d','MarkerIndices',1:50:length(tpar_N),'MarkerSize',4,'color',[17 17 17]./255);
h =[p1;p2;p3];

legendBubble{iBubble,1} =['Simulation'];
legendBubble{iBubble,2} =['Analytic, Bubble No.',int2str(iBubble)];
legendBubble{iBubble,3} =['Semi Analytic, Bubble No.',int2str(iBubble)];
legend(h,legendBubble{iBubble,:},'Location','northwest')
title(['Analytic vs Simulation of MB scattered pressure at (x,y,z) = (',...
    num2str(round(100*domain.par{domain.dimval(2)}(SliceIndex(1)))/100),', ',...
    num2str(round(100*domain.par{domain.dimval(1)}(SliceIndex(2)))/100),', '...
    num2str(round(100*domain.par{domain.dimval(3)}(SliceIndex(3)))/100),') [mm] , iteration ',num2str(i),', contribution of ',num2str(iBubble),' Scatterer(s)'])
xlim([tpar_N(1) tpar_N(end)]*1E6)
RRMS_semi_an = ['RRMS (Semi) ='        ,num2str(round(RRMS_ErrorG_semi_an(k,m) *1000)/10),'[%]'];
% RelE_semi_an = ['|Rel.Error| (Semi) = ',num2str(round(Rel_ErrorG_semi_an*1000)/10),'[%]'];
RRMS_an      = ['RRMS (Full) ='        ,num2str(round(RRMS_ErrorG_an(k,m) *1000)/10),'[%]'];
% RelE_an      = ['|Rel.Error| (Full) = ',num2str(round(Rel_ErrorG_an*1000)/10),'[%]'];
annotation('textbox', [0.75, 0.81, 0.1, 0.1], 'String', [RRMS_an newline RRMS_semi_an])
xlabel('Time [usec]')
ylabel('Pressure [Pa]')
set(gca,'FontSize',20)
set(gcf,'color','white')
grid on;

end