function generate_ICS_configurationfile(filename,mediumparams,domainparams,...
    modelparams,sourceparams)

% generate_ICS_configurationfile(filename,mediumparams,...
%   domainparams,modelparams,sourceparams)
% Generates the configuration file by the ICS Parnac program.
% Inputs:
% filename:     filename of the configuration file. Do not include path.
% mediumparams: structure that has been generated with set_ICS_medium.
% domainparams: structure that has been generated with set_ICS_domain.
% modelparams:  structure that has been generated with set_ICS_model. 
% sourceparams: structure that has been generated with set_ICS_source_parametric
%               or with set_ICS_source_data.
%
% The configuration file version generated by this function is 1.01

% JH, version 090116

if(nargin<1)
    error('Not enough arguments')
end
if(nargin>5)
    error('Too many arguments')
end

% Check and correct several parameters which can only be done in the 
% combination of all structures.
% No further input argument checking, it is assumed that all structures
% have been generated with the appropriate set_ICS_ functions and have been
% corrected and checked there.

% The fundamental frequency in modelparams is the leading one.
freq0 = modelparams.freq0;

% if fc0 is not defined, use modelparams.freq0 instead
if(mediumparams.fc0==-1)
    fc0 = freq0;
else
    fc0 = mediumparams.fc0;
end

% The freq0 in the modelparams structure is the leading fundamental frequency,
% other freq0 settings are compared with this one and give an error if incompatible
[lt,st] = normalize_lt_st_to_freq0(domainparams,freq0);

% - account for st in lt, Tpulse and in Tdelay
%   the employment of st has not been implemented in the Fortran program,
%   but accounting for it in lt, Tpulse and Tdelay provides an easy
%   bypass...
lt = lt + st;
if(strcmp(sourceparams.definitiontype,'parametric')...
        &&~strcmp(sourceparams.signature.signaturetype,'file'))
        Tpulse = sourceparams.signature.Tpulse + st;
        Tdelay = sourceparams.signature.Tdelay + st;
else
    if(st~=0)
        % If the source or the signature is defined as a data file, then the extra
        % delay introduced by st has to be included in the files themselves.
        disp('Warning: source or signature is defined as a data file and st~=0. Account for st as an extra delay in the data file');
    end
end

%write configuration file
fid=fopen(filename,'w');

% Remark: Fortran uses only \n as end-of-line character
% Windows would require \r\n as end-of-line
fprintf(fid,'*** Parnac Input File ***\n');
fprintf(fid,'*************************\n');
fprintf(fid,'1.01			 *** Version number\n');
fprintf(fid,'***************************\n');
fprintf(fid,'***  medium parameters  ***\n');
fprintf(fid,'***************************\n');
fprintf(fid,'%6.6E		*** rho0   = Ambient density rho0 [kg/m^3]\n',mediumparams.rho0);
fprintf(fid,'%6.6E		*** c0     = Small signal sound velocity [m/s]\n',mediumparams.c0);
fprintf(fid,'%6.6E		*** fc0    = Frequency at which sound velocity is given [Hz]\n',fc0);
fprintf(fid,'%6.6E		*** beta   = Nonlinearity parameter\n',mediumparams.beta);
fprintf(fid,'%6.6E		*** a_att  = Frequency-dependent attenuation parameter [Np/cm MHz^b]\n',mediumparams.a_att);
fprintf(fid,'%6.6E		*** b_att  = Power in frequency power attenuation law\n',mediumparams.b_att);
fprintf(fid,'*************************************************\n');
fprintf(fid,'***   domain and discretization parameters    ***\n');
fprintf(fid,'*************************************************\n');
fprintf(fid,'%6.6E		*** Lx in [mm]\n',domainparams.lx);
fprintf(fid,'%6.6E		*** Ly in [mm]\n',domainparams.ly);
fprintf(fid,'%6.6E		*** Lt in [cycles]\n',lt);
fprintf(fid,'%6.6E		*** Sz in [mm]	(Start of the field)\n',domainparams.sz);
fprintf(fid,'%6.6E		*** Lz in [mm]	(length of the field in z)\n',domainparams.lz);
fprintf(fid,'%6.6E		*** Thetax [rad]\n',domainparams.thetax);
fprintf(fid,'%6.6E		*** Thetay [rad]\n',domainparams.thetay);
fprintf(fid,'%6.6E		*** Thetat [rad]\n',domainparams.thetat);
fprintf(fid,'%6.6E		*** freq0  = Fundamental frequency [Hz]\n',freq0);
fprintf(fid,'%6.6E		*** Fnyq   = Maximum frequency of interest (in fundamental freq)\n',modelparams.Fnyq);
fprintf(fid,'************************************\n');
fprintf(fid,'***    other model parameters    ***\n');
fprintf(fid,'************************************\n');
fprintf(fid,'*** PrimarySourceType, choices: "qsource", "fsource", "dtqsource", "planewave"\n');
fprintf(fid,[sourceparams.excitationtype '\n']);
fprintf(fid,'*** Contrast source type, choices: "nonlin", "complexcontrast", "sphere", "luneberg", "blob"\n');
fprintf(fid,[modelparams.contrastsourcetype '\n']);
fprintf(fid,'%1.0f		*** Order of the Temporal Finite Difference schemes\n',modelparams.FDTorder);
fprintf(fid,'%1.0f		*** Order of the Spatial  Finite Difference schemes\n',modelparams.FDXorder);
fprintf(fid,'%1.0f		*** Debug message level 0-5 gives the level of detail of the messages\n',modelparams.debuglevel);
if(~modelparams.useantialiasing)
    fprintf(fid,'0		*** Switch: Use Anti-Aliasing in contrast source evaluation\n');
else
    fprintf(fid,'1		*** Switch: Use Anti-Aliasing in contrast source evaluation\n');
end
if(~modelparams.usesupporttapering)
    fprintf(fid,'0		*** Switch: Use Support Tapering in contrast source evaluation	\n');
else
    fprintf(fid,'1		*** Switch: Use Support Tapering in contrast source evaluation	\n');
end
if(~modelparams.usefreqtapering)
    fprintf(fid,'0		*** Switch: Use MaxFreq Tapering in contrast source evaluation \n');
else
    fprintf(fid,'1		*** Switch: Use MaxFreq Tapering in contrast source evaluation \n');
end
fprintf(fid,'%1.0f		*** Number of beams in which the total field is split\n',numel(modelparams.beamiterations));
fprintf(fid,'%1.0f		*** Number of iterations of the first beam\n',modelparams.beamiterations(1));
if(numel(modelparams.beamiterations)>1)
    for i=2:numel(modelparams.beamiterations)-1
        fprintf(fid,'%1.0f		*** Number of iterations of the n-th beam\n',modelparams.beamiterations(i));
    end
    fprintf(fid,'%1.0f		*** Number of iterations of the last beam\n',modelparams.beamiterations(numel(modelparams.beamiterations)));
end
fprintf(fid,'*** Save slices, choices: "firstandlast", "last", "all"\n');
fprintf(fid,[modelparams.slicesavetype '\n']);
fprintf(fid,'%1.0f		*** Number of slices to be stored\n',numel(modelparams.slicesavepos));
for i=1:numel(modelparams.slicesavepos)
    fprintf(fid,[ modelparams.slicesavedim(i) ' %6.6E		*** Slice x/y/z, position in [mm] (skew beams: position in x/y only at z=0)\n'],modelparams.slicesavepos(i));
end
fprintf(fid,'***************************\n');
fprintf(fid,'***  source parameters  ***\n');
fprintf(fid,'***************************\n');
fprintf(fid,'***   source inputfilename    *** Use "standard" for [inputfilename]_src.mat, use "none" for parametric source definition (below)\n');
if(strcmp(sourceparams.definitiontype,'file'))
    fprintf(fid,[sourceparams.filename '\n']);
elseif(strcmp(sourceparams.definitiontype,'parametric'))
    fprintf(fid,'none\n');
    fprintf(fid,'*** Source signature, choices: "gaussian", "blackman" or filename\n');
    if(strcmp(sourceparams.signature.signaturetype,'file'))
        fprintf(fid,[sourceparams.signature.signaturefilename '\n']);
    elseif(strcmp(sourceparams.signature.signaturetype,'gaussian'))
        fprintf(fid,[sourceparams.signature.signaturetype '\n']);
        fprintf(fid,'%6.6E		*** Pstart = Pressure at source surface [Pa] - including 2*pi*f0 for source type 2\n',sourceparams.signature.Pstart);
        fprintf(fid,'%6.6E		*** Tpulse = Length of pulse (in 1/(fundamental freq)) to take into account in calculations\n',Tpulse);
        fprintf(fid,'*** Gaussian specific : Twidth (in 1/freq0), Tdelay (in 1/freq0), Power (integer)\n');
        fprintf(fid,'*** Blackman specific : Tdelay (in 1/freq0)\n');
        fprintf(fid,'%6.6E		*** Twidth = Width of gaussian window\n',sourceparams.signature.Twidth);
        fprintf(fid,'%6.6E		*** Tdelay = Delay of gaussian window\n',Tdelay);
        fprintf(fid,'%1.0f		*** Power  = Power of gaussian window	\n',sourceparams.signature.power);
    elseif(strcmp(sourceparams.signature.signaturetype,'blackman'))
        fprintf(fid,[sourceparams.signature.signaturetype '\n']);
        fprintf(fid,'%6.6E		*** Pstart = Pressure at source surface [Pa] - including 2*pi*f0 for source type 2\n',sourceparams.signature.Pstart);
        fprintf(fid,'%6.6E		*** Tpulse = Length of pulse (in 1/(fundamental freq)) to take into account in calculations\n',Tpulse);
        fprintf(fid,'*** Gaussian specific : Twidth (in 1/4*freq0), Tdelay (in 1/freq0), Power (integer)\n');
        fprintf(fid,'*** Blackman specific : Tdelay (in 1/freq0)\n');
        fprintf(fid,'%6.6E		*** Tdelay = Delay of gaussian window\n',Tdelay);
    end
    if(~strcmp(sourceparams.excitationtype,'planewave'))
        fprintf(fid,'*** Source shape, choices: "cylindricalflat", "rectangularflat", "triangularflat", "pointsource", "phasedarray"\n');
        fprintf(fid,[sourceparams.aperture.aperturetype '\n']);
        fprintf(fid,'*** cylindricalflat specific: Radius (in mm), Radfocus (in mm)\n');
        fprintf(fid,'*** rectangularflat specific: Width(x, in mm), Height(y, in mm)\n');
        fprintf(fid,'*** triangularflat specific: Edge size (in mm), Position of the center in the x-dimension (in mm) (top corner of the triangle lies on the positive x-axis)\n');
        fprintf(fid,'*** pointsource specific: No parameters\n');
        fprintf(fid,'*** phasedarray specific: Number of Elements NEl, El. Width(x, in mm), El. Height(y, in mm), kerf(x, in mm), Focus X (mm), Focus Z (mm), Elevation Focus Z (mm), PhaseApodFilename\n');
        if(strcmp(sourceparams.aperture.aperturetype,'cylindrical'))
            fprintf(fid,'%6.6E		*** Radius\n',sourceparams.aperture.radius);
            fprintf(fid,'%6.6E		*** Radfocus\n',sourceparams.aperture.radfocus);
        elseif(strcmp(sourceparams.aperture.aperturetype,'rectangular'))
            fprintf(fid,'%6.6E		*** Rectwidth\n',sourceparams.aperture.rectwidth);
            fprintf(fid,'%6.6E		*** Rectheight\n',sourceparams.aperture.rectheight);
        elseif(strcmp(sourceparams.aperture.aperturetype,'triangular'))
            fprintf(fid,'%6.6E		*** Triangedge\n',sourceparams.aperture.triangedge);
            fprintf(fid,'%6.6E		*** Triangxcenter\n',sourceparams.aperture.triangxcenter);
        elseif(strcmp(sourceparams.aperture.aperturetype,'pointsource'))
        elseif(strcmp(sourceparams.aperture.aperturetype,'phasedarray'))
            fprintf(fid,'%1.0f		*** Number of elements\n',sourceparams.aperture.numels);
            fprintf(fid,'%6.6E		*** Element width\n',sourceparams.aperture.elwidth);
            fprintf(fid,'%6.6E		*** Element height\n',sourceparams.aperture.elheight);
            fprintf(fid,'%6.6E		*** Kerf\n',sourceparams.aperture.kerf);
            fprintf(fid,'%6.6E		*** Focus x\n',sourceparams.aperture.focusx);
            fprintf(fid,'%6.6E		*** Focus z\n',sourceparams.aperture.focusz);
            fprintf(fid,'%6.6E		*** Elevation focus\n',sourceparams.aperture.elevationfocusz);
            fprintf(fid,'***   apodization inputfilename    *** Use "standard" for [inputfilename]_apo.mat\n');
            fprintf(fid,[sourceparams.aperture.apofilename '\n']);
        end
    end
end

fclose(fid);


